#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import os
import sys
import time
import logging
from threading import Condition
try:
    import thread
except ImportError:
    # py 3
    import _thread as thread
try:
    from Queue import Queue
except ImportError:
    # py 3
    from queue import Queue
try:
    import cPickle as pickle
except ImportError:
    # py 3
    import pickle

from syncdutils import Thread, select

pickle_proto = -1
repce_version = 1.0


def ioparse(i, o):
    if isinstance(i, int):
        i = os.fdopen(i)
    # rely on duck typing for recognizing
    # streams as that works uniformly
    # in py2 and py3
    if hasattr(o, 'fileno'):
        o = o.fileno()
    return (i, o)


def send(out, *args):
    """pickle args and write out wholly in one syscall

    ie. not use the ability of pickle to dump directly to
    a stream, as that would potentially mess up messages
    by interleaving them
    """
    os.write(out, pickle.dumps(args, pickle_proto))


def recv(inf):
    """load an object from input stream"""
    return pickle.load(inf)


class RepceServer(object):

    """RePCe is Hungarian for canola, http://hu.wikipedia.org/wiki/Repce

    ... also our homebrewed RPC backend where the transport layer is
    reduced to a pair of filehandles.

    This is the server component.
    """

    def __init__(self, obj, i, o, wnum=6):
        """register a backend object .obj to which incoming messages
           are dispatched, also incoming/outcoming streams
        """
        self.obj = obj
        self.inf, self.out = ioparse(i, o)
        self.wnum = wnum
        self.q = Queue()

    def service_loop(self):
        """fire up worker threads, get messages and dispatch among them"""
        for i in range(self.wnum):
            t = Thread(target=self.worker)
            t.start()
        try:
            while True:
                self.q.put(recv(self.inf))
        except EOFError:
            logging.info("terminating on reaching EOF.")

    def worker(self):
        """life of a worker

        Get message, extract its id, method name and arguments
        (kwargs not supported), call method on .obj.
        Send back message id + return value.
        If method call throws an exception, rescue it, and send
        back the exception as result (with flag marking it as
        exception).
        """
        while True:
            in_data = self.q.get(True)
            rid = in_data[0]
            rmeth = in_data[1]
            exc = False
            if rmeth == '__repce_version__':
                res = repce_version
            else:
                try:
                    res = getattr(self.obj, rmeth)(*in_data[2:])
                except:
                    res = sys.exc_info()[1]
                    exc = True
                    logging.exception("call failed: ")
            send(self.out, rid, exc, res)


class RepceJob(object):

    """class representing message status we can use
    for waiting on reply"""

    def __init__(self, cbk):
        """
        - .rid: (process-wise) unique id
        - .cbk: what we do upon receiving reply
        """
        self.rid = (os.getpid(), thread.get_ident(), time.time())
        self.cbk = cbk
        self.lever = Condition()
        self.done = False

    def __repr__(self):
        return ':'.join([str(x) for x in self.rid])

    def wait(self):
        self.lever.acquire()
        if not self.done:
            self.lever.wait()
        self.lever.release()
        return self.result

    def wakeup(self, data):
        self.result = data
        self.lever.acquire()
        self.done = True
        self.lever.notify()
        self.lever.release()


class RepceClient(object):

    """RePCe is Hungarian for canola, http://hu.wikipedia.org/wiki/Repce

    ... also our homebrewed RPC backend where the transport layer is
    reduced to a pair of filehandles.

    This is the client component.
    """

    def __init__(self, i, o):
        self.inf, self.out = ioparse(i, o)
        self.jtab = {}
        t = Thread(target=self.listen)
        t.start()

    def listen(self):
        while True:
            select((self.inf,), (), ())
            rid, exc, res = recv(self.inf)
            rjob = self.jtab.pop(rid)
            if rjob.cbk:
                rjob.cbk(rjob, [exc, res])

    def push(self, meth, *args, **kw):
        """wrap arguments in a RepceJob, send them to server
           and return the RepceJob

           @cbk to pass on RepceJob can be given as kwarg.
        """
        cbk = kw.get('cbk')
        if not cbk:
            def cbk(rj, res):
                if res[0]:
                    raise res[1]
        rjob = RepceJob(cbk)
        self.jtab[rjob.rid] = rjob
        logging.debug("call %s %s%s ..." % (repr(rjob), meth, repr(args)))
        send(self.out, rjob.rid, meth, *args)
        return rjob

    def __call__(self, meth, *args):
        """RePCe client is callabe, calling it implements a synchronous
        remote call.

        We do a .push with a cbk which does a wakeup upon receiving anwser,
        then wait on the RepceJob.
        """
        rjob = self.push(
            meth, *args, **{'cbk': lambda rj, res: rj.wakeup(res)})
        exc, res = rjob.wait()
        if exc:
            logging.error('call %s (%s) failed on peer with %s' %
                          (repr(rjob), meth, str(type(res).__name__)))
            raise res
        logging.debug("call %s %s -> %s" % (repr(rjob), meth, repr(res)))
        return res

    class mprx(object):

        """method proxy, standard trick to implement rubyesque
        method_missing in Python

        A class is a closure factory, you know what I mean, or go read
        some SICP.
        """

        def __init__(self, ins, meth):
            self.ins = ins
            self.meth = meth

        def __call__(self, *a):
            return self.ins(self.meth, *a)

    def __getattr__(self, meth):
        """this implements transparent method dispatch to remote object,
           so that you don't need to call the RepceClient instance like

             rclient('how_old_are_you_if_born_in', 1979)

           but you can make it into an ordinary method call like

             rclient.how_old_are_you_if_born_in(1979)
        """
        return self.mprx(self, meth)

    def __version__(self):
        """used in handshake to verify compatibility"""
        d = {'proto': self('__repce_version__')}
        try:
            d['object'] = self('version')
        except AttributeError:
            pass
        return d



#!/usr/bin/env python
#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import os
import os.path
import glob
import sys
import time
import logging
import shutil
import optparse
import fcntl
import fnmatch
from optparse import OptionParser, SUPPRESS_HELP
from logging import Logger, handlers
from errno import ENOENT

from ipaddr import IPAddress, IPNetwork

from gconf import gconf
from syncdutils import FreeObject, norm, grabpidfile, finalize
from syncdutils import log_raise_exception, privileged
from syncdutils import GsyncdError, select, set_term_handler
from configinterface import GConffile, upgrade_config_file
import resource
from monitor import monitor
from changelogagent import agent, Changelog
from gsyncdstatus import set_monitor_status, GeorepStatus


class GLogger(Logger):

    """Logger customizations for gsyncd.

    It implements a log format similar to that of glusterfs.
    """

    def makeRecord(self, name, level, *a):
        rv = Logger.makeRecord(self, name, level, *a)
        rv.nsecs = (rv.created - int(rv.created)) * 1000000
        fr = sys._getframe(4)
        callee = fr.f_locals.get('self')
        if callee:
            ctx = str(type(callee)).split("'")[1].split('.')[-1]
        else:
            ctx = '<top>'
        if not hasattr(rv, 'funcName'):
            rv.funcName = fr.f_code.co_name
        rv.lvlnam = logging.getLevelName(level)[0]
        rv.ctx = ctx
        return rv

    @classmethod
    def setup(cls, **kw):
        lbl = kw.get('label', "")
        if lbl:
            lbl = '(' + lbl + ')'
        lprm = {'datefmt': "%Y-%m-%d %H:%M:%S",
                'format': "[%(asctime)s.%(nsecs)d] %(lvlnam)s [%(module)s" +
                lbl + ":%(lineno)s:%(funcName)s] %(ctx)s: %(message)s"}
        lprm.update(kw)
        lvl = kw.get('level', logging.INFO)
        lprm['level'] = lvl
        logging.root = cls("root", lvl)
        logging.setLoggerClass(cls)
        logging.getLogger().handlers = []
        logging.getLogger().setLevel(lprm['level'])

        if 'filename' in lprm:
            try:
                logging_handler = handlers.WatchedFileHandler(lprm['filename'])
                formatter = logging.Formatter(fmt=lprm['format'],
                                              datefmt=lprm['datefmt'])
                logging_handler.setFormatter(formatter)
                logging.getLogger().addHandler(logging_handler)
            except AttributeError:
                # Python version < 2.6 will not have WatchedFileHandler
                # so fallback to logging without any handler.
                # Note: logrotate will not work if Python version is < 2.6
                logging.basicConfig(**lprm)
        else:
            # If filename not passed(not available in lprm) then it may be
            # streaming.(Ex: {"stream": "/dev/stdout"})
            logging.basicConfig(**lprm)

    @classmethod
    def _gsyncd_loginit(cls, **kw):
        lkw = {}
        if gconf.log_level:
            lkw['level'] = gconf.log_level
        if kw.get('log_file'):
            if kw['log_file'] in ('-', '/dev/stderr'):
                lkw['stream'] = sys.stderr
            elif kw['log_file'] == '/dev/stdout':
                lkw['stream'] = sys.stdout
            else:
                lkw['filename'] = kw['log_file']

        cls.setup(label=kw.get('label'), **lkw)

        lkw.update({'saved_label': kw.get('label')})
        gconf.log_metadata = lkw
        gconf.log_exit = True


def startup(**kw):
    """set up logging, pidfile grabbing, daemonization"""
    if getattr(gconf, 'pid_file', None) and kw.get('go_daemon') != 'postconn':
        if not grabpidfile():
            sys.stderr.write("pidfile is taken, exiting.\n")
            sys.exit(2)
        gconf.pid_file_owned = True

    if kw.get('go_daemon') == 'should':
        x, y = os.pipe()
        gconf.cpid = os.fork()
        if gconf.cpid:
            os.close(x)
            sys.exit()
        os.close(y)
        os.setsid()
        dn = os.open(os.devnull, os.O_RDWR)
        for f in (sys.stdin, sys.stdout, sys.stderr):
            os.dup2(dn, f.fileno())
        if getattr(gconf, 'pid_file', None):
            if not grabpidfile(gconf.pid_file + '.tmp'):
                raise GsyncdError("cannot grab temporary pidfile")
            os.rename(gconf.pid_file + '.tmp', gconf.pid_file)
        # wait for parent to terminate
        # so we can start up with
        # no messing from the dirty
        # ol' bustard
        select((x,), (), ())
        os.close(x)

    GLogger._gsyncd_loginit(**kw)


def _unlink(path):
    try:
        os.unlink(path)
    except (OSError, IOError):
        if sys.exc_info()[1].errno == ENOENT:
            pass
        else:
            raise GsyncdError('Unlink error: %s' % path)


def main():
    """main routine, signal/exception handling boilerplates"""
    gconf.starttime = time.time()
    set_term_handler()
    GLogger.setup()
    excont = FreeObject(exval=0)
    try:
        try:
            main_i()
        except:
            log_raise_exception(excont)
    finally:
        finalize(exval=excont.exval)


def main_i():
    """internal main routine

    parse command line, decide what action will be taken;
    we can either:
    - query/manipulate configuration
    - format gsyncd urls using gsyncd's url parsing engine
    - start service in following modes, in given stages:
      - agent: startup(), ChangelogAgent()
      - monitor: startup(), monitor()
      - master: startup(), connect_remote(), connect(), service_loop()
      - slave: startup(), connect(), service_loop()
    """
    rconf = {'go_daemon': 'should'}

    def store_abs(opt, optstr, val, parser):
        if val and val != '-':
            val = os.path.abspath(val)
        setattr(parser.values, opt.dest, val)

    def store_local(opt, optstr, val, parser):
        rconf[opt.dest] = val

    def store_local_curry(val):
        return lambda o, oo, vx, p: store_local(o, oo, val, p)

    def store_local_obj(op, dmake):
        return lambda o, oo, vx, p: store_local(
            o, oo, FreeObject(op=op, **dmake(vx)), p)

    op = OptionParser(
        usage="%prog [options...] <master> <slave>", version="%prog 0.0.1")
    op.add_option('--gluster-command-dir', metavar='DIR', default='')
    op.add_option('--gluster-log-file', metavar='LOGF',
                  default=os.devnull, type=str, action='callback',
                  callback=store_abs)
    op.add_option('--gluster-log-level', metavar='LVL')
    op.add_option('--gluster-params', metavar='PRMS', default='')
    op.add_option(
        '--glusterd-uuid', metavar='UUID', type=str, default='',
        help=SUPPRESS_HELP)
    op.add_option(
        '--gluster-cli-options', metavar='OPTS', default='--log-file=-')
    op.add_option('--mountbroker', metavar='LABEL')
    op.add_option('-p', '--pid-file', metavar='PIDF', type=str,
                  action='callback', callback=store_abs)
    op.add_option('-l', '--log-file', metavar='LOGF', type=str,
                  action='callback', callback=store_abs)
    op.add_option('--iprefix',  metavar='LOGD',  type=str,
                  action='callback', callback=store_abs)
    op.add_option('--changelog-log-file',  metavar='LOGF',  type=str,
                  action='callback', callback=store_abs)
    op.add_option('--log-file-mbr', metavar='LOGF', type=str,
                  action='callback', callback=store_abs)
    op.add_option('--state-file', metavar='STATF', type=str,
                  action='callback', callback=store_abs)
    op.add_option('--state-detail-file', metavar='STATF',
                  type=str, action='callback', callback=store_abs)
    op.add_option('--georep-session-working-dir', metavar='STATF',
                  type=str, action='callback', callback=store_abs)
    op.add_option('--ignore-deletes', default=False, action='store_true')
    op.add_option('--isolated-slave', default=False, action='store_true')
    op.add_option('--use-rsync-xattrs', default=False, action='store_true')
    op.add_option('--sync-xattrs', default=True, action='store_true')
    op.add_option('--sync-acls', default=True, action='store_true')
    op.add_option('--log-rsync-performance', default=False,
                  action='store_true')
    op.add_option('--pause-on-start', default=False, action='store_true')
    op.add_option('-L', '--log-level', metavar='LVL')
    op.add_option('-r', '--remote-gsyncd', metavar='CMD',
                  default=os.path.abspath(sys.argv[0]))
    op.add_option('--volume-id', metavar='UUID')
    op.add_option('--slave-id', metavar='ID')
    op.add_option('--session-owner', metavar='ID')
    op.add_option('--local-id', metavar='ID', help=SUPPRESS_HELP, default='')
    op.add_option(
        '--local-path', metavar='PATH', help=SUPPRESS_HELP, default='')
    op.add_option('-s', '--ssh-command', metavar='CMD', default='ssh')
    op.add_option('--ssh-command-tar', metavar='CMD', default='ssh')
    op.add_option('--rsync-command', metavar='CMD', default='rsync')
    op.add_option('--rsync-options', metavar='OPTS', default='')
    op.add_option('--rsync-ssh-options', metavar='OPTS', default='--compress')
    op.add_option('--timeout', metavar='SEC', type=int, default=120)
    op.add_option('--connection-timeout', metavar='SEC',
                  type=int, default=60, help=SUPPRESS_HELP)
    op.add_option('--sync-jobs', metavar='N', type=int, default=3)
    op.add_option('--replica-failover-interval', metavar='N',
                  type=int, default=1)
    op.add_option('--changelog-archive-format', metavar='N',
                  type=str, default="%Y%m")
    op.add_option('--use-meta-volume', default=False, action='store_true')
    op.add_option('--meta-volume-mnt', metavar='N',
                  type=str, default="/var/run/gluster/shared_storage")
    op.add_option(
        '--turns', metavar='N', type=int, default=0, help=SUPPRESS_HELP)
    op.add_option('--allow-network', metavar='IPS', default='')
    op.add_option('--socketdir', metavar='DIR')
    op.add_option('--state-socket-unencoded', metavar='SOCKF',
                  type=str, action='callback', callback=store_abs)
    op.add_option('--checkpoint', metavar='LABEL', default='0')

    # tunables for failover/failback mechanism:
    # None   - gsyncd behaves as normal
    # blind  - gsyncd works with xtime pairs to identify
    #          candidates for synchronization
    # wrapup - same as normal mode but does not assign
    #          xtimes to orphaned files
    # see crawl() for usage of the above tunables
    op.add_option('--special-sync-mode', type=str, help=SUPPRESS_HELP)

    # changelog or xtime? (TODO: Change the default)
    op.add_option(
        '--change-detector', metavar='MODE', type=str, default='xtime')
    # sleep interval for change detection (xtime crawl uses a hardcoded 1
    # second sleep time)
    op.add_option('--change-interval', metavar='SEC', type=int, default=3)
    # working directory for changelog based mechanism
    op.add_option('--working-dir', metavar='DIR', type=str,
                  action='callback', callback=store_abs)
    op.add_option('--use-tarssh', default=False, action='store_true')

    op.add_option('-c', '--config-file', metavar='CONF',
                  type=str, action='callback', callback=store_local)
    # duh. need to specify dest or value will be mapped to None :S
    op.add_option('--monitor', dest='monitor', action='callback',
                  callback=store_local_curry(True))
    op.add_option('--agent', dest='agent', action='callback',
                  callback=store_local_curry(True))
    op.add_option('--resource-local', dest='resource_local',
                  type=str, action='callback', callback=store_local)
    op.add_option('--resource-remote', dest='resource_remote',
                  type=str, action='callback', callback=store_local)
    op.add_option('--feedback-fd', dest='feedback_fd', type=int,
                  help=SUPPRESS_HELP, action='callback', callback=store_local)
    op.add_option('--rpc-fd', dest='rpc_fd', type=str, help=SUPPRESS_HELP)
    op.add_option('--subvol-num', dest='subvol_num', type=int,
                  help=SUPPRESS_HELP)
    op.add_option('--listen', dest='listen', help=SUPPRESS_HELP,
                  action='callback', callback=store_local_curry(True))
    op.add_option('-N', '--no-daemon', dest="go_daemon",
                  action='callback', callback=store_local_curry('dont'))
    op.add_option('--verify', type=str, dest="verify",
                  action='callback', callback=store_local)
    op.add_option('--create', type=str, dest="create",
                  action='callback', callback=store_local)
    op.add_option('--delete', dest='delete', action='callback',
                  callback=store_local_curry(True))
    op.add_option('--status-get', dest='status_get', action='callback',
                  callback=store_local_curry(True))
    op.add_option('--debug', dest="go_daemon", action='callback',
                  callback=lambda *a: (store_local_curry('dont')(*a),
                                       setattr(
                                           a[-1].values, 'log_file', '-'),
                                       setattr(a[-1].values, 'log_level',
                                               'DEBUG'),
                                       setattr(a[-1].values,
                                               'changelog_log_file', '-')))
    op.add_option('--path', type=str, action='append')

    for a in ('check', 'get'):
        op.add_option('--config-' + a, metavar='OPT', type=str, dest='config',
                      action='callback',
                      callback=store_local_obj(a, lambda vx: {'opt': vx}))
    op.add_option('--config-get-all', dest='config', action='callback',
                  callback=store_local_obj('get', lambda vx: {'opt': None}))
    for m in ('', '-rx', '-glob'):
        # call this code 'Pythonic' eh?
        # have to define a one-shot local function to be able
        # to inject (a value depending on the)
        # iteration variable into the inner lambda
        def conf_mod_opt_regex_variant(rx):
            op.add_option('--config-set' + m, metavar='OPT VAL', type=str,
                          nargs=2, dest='config', action='callback',
                          callback=store_local_obj('set', lambda vx: {
                              'opt': vx[0], 'val': vx[1], 'rx': rx}))
            op.add_option('--config-del' + m, metavar='OPT', type=str,
                          dest='config', action='callback',
                          callback=store_local_obj('del', lambda vx: {
                              'opt': vx, 'rx': rx}))
        conf_mod_opt_regex_variant(m and m[1:] or False)

    op.add_option('--normalize-url', dest='url_print',
                  action='callback', callback=store_local_curry('normal'))
    op.add_option('--canonicalize-url', dest='url_print',
                  action='callback', callback=store_local_curry('canon'))
    op.add_option('--canonicalize-escape-url', dest='url_print',
                  action='callback', callback=store_local_curry('canon_esc'))

    tunables = [norm(o.get_opt_string()[2:])
                for o in op.option_list
                if (o.callback in (store_abs, 'store_true', None) and
                    o.get_opt_string() not in ('--version', '--help'))]
    remote_tunables = ['listen', 'go_daemon', 'timeout',
                       'session_owner', 'config_file', 'use_rsync_xattrs']
    rq_remote_tunables = {'listen': True}

    # precedence for sources of values: 1) commandline, 2) cfg file, 3)
    # defaults for this to work out we need to tell apart defaults from
    # explicitly set options... so churn out the defaults here and call
    # the parser with virgin values container.
    defaults = op.get_default_values()
    opts, args = op.parse_args(values=optparse.Values())
    args_orig = args[:]
    r = rconf.get('resource_local')
    if r:
        if len(args) == 0:
            args.append(None)
        args[0] = r
    r = rconf.get('resource_remote')
    if r:
        if len(args) == 0:
            raise GsyncdError('local resource unspecfied')
        elif len(args) == 1:
            args.append(None)
        args[1] = r
    confdata = rconf.get('config')
    if not (len(args) == 2 or
            (len(args) == 1 and rconf.get('listen')) or
            (len(args) <= 2 and confdata) or
            rconf.get('url_print')):
        sys.stderr.write("error: incorrect number of arguments\n\n")
        sys.stderr.write(op.get_usage() + "\n")
        sys.exit(1)

    verify = rconf.get('verify')
    if verify:
        logging.info(verify)
        logging.info("Able to spawn gsyncd.py")
        return

    restricted = os.getenv('_GSYNCD_RESTRICTED_')

    if restricted:
        allopts = {}
        allopts.update(opts.__dict__)
        allopts.update(rconf)
        bannedtuns = set(allopts.keys()) - set(remote_tunables)
        if bannedtuns:
            raise GsyncdError('following tunables cannot be set with '
                              'restricted SSH invocaton: ' +
                              ', '.join(bannedtuns))
        for k, v in rq_remote_tunables.items():
            if not k in allopts or allopts[k] != v:
                raise GsyncdError('tunable %s is not set to value %s required '
                                  'for restricted SSH invocaton' %
                                  (k, v))

    confrx = getattr(confdata, 'rx', None)

    def makersc(aa, check=True):
        if not aa:
            return ([], None, None)
        ra = [resource.parse_url(u) for u in aa]
        local = ra[0]
        remote = None
        if len(ra) > 1:
            remote = ra[1]
        if check and not local.can_connect_to(remote):
            raise GsyncdError("%s cannot work with %s" %
                              (local.path, remote and remote.path))
        return (ra, local, remote)
    if confrx:
        # peers are regexen, don't try to parse them
        if confrx == 'glob':
            args = ['\A' + fnmatch.translate(a) for a in args]
        canon_peers = args
        namedict = {}
    else:
        dc = rconf.get('url_print')
        rscs, local, remote = makersc(args_orig, not dc)
        if dc:
            for r in rscs:
                print(r.get_url(**{'normal': {},
                                   'canon': {'canonical': True},
                                   'canon_esc': {'canonical': True,
                                                 'escaped': True}}[dc]))
            return
        pa = ([], [], [])
        urlprms = (
            {}, {'canonical': True}, {'canonical': True, 'escaped': True})
        for x in rscs:
            for i in range(len(pa)):
                pa[i].append(x.get_url(**urlprms[i]))
        _, canon_peers, canon_esc_peers = pa
        # creating the namedict, a dict representing various ways of referring
        # to / repreenting peers to be fillable in config templates
        mods = (lambda x: x, lambda x: x[
                0].upper() + x[1:], lambda x: 'e' + x[0].upper() + x[1:])
        if remote:
            rmap = {local: ('local', 'master'), remote: ('remote', 'slave')}
        else:
            rmap = {local: ('local', 'slave')}
        namedict = {}
        for i in range(len(rscs)):
            x = rscs[i]
            for name in rmap[x]:
                for j in range(3):
                    namedict[mods[j](name)] = pa[j][i]
                namedict[name + 'vol'] = x.volume
                if name == 'remote':
                    namedict['remotehost'] = x.remotehost
    if not 'config_file' in rconf:
        rconf['config_file'] = os.path.join(
            os.path.dirname(sys.argv[0]), "conf/gsyncd_template.conf")

    upgrade_config_file(rconf['config_file'])
    gcnf = GConffile(
        rconf['config_file'], canon_peers,
        defaults.__dict__, opts.__dict__, namedict)

    checkpoint_change = False
    if confdata:
        opt_ok = norm(confdata.opt) in tunables + [None]
        if confdata.op == 'check':
            if opt_ok:
                sys.exit(0)
            else:
                sys.exit(1)
        elif not opt_ok:
            raise GsyncdError("not a valid option: " + confdata.opt)
        if confdata.op == 'get':
            gcnf.get(confdata.opt)
        elif confdata.op == 'set':
            gcnf.set(confdata.opt, confdata.val, confdata.rx)
        elif confdata.op == 'del':
            gcnf.delete(confdata.opt, confdata.rx)
        # when modifying checkpoint, it's important to make a log
        # of that, so in that case we go on to set up logging even
        # if its just config invocation
        if confdata.opt == 'checkpoint' and confdata.op in ('set', 'del') and \
           not confdata.rx:
            checkpoint_change = True
        if not checkpoint_change:
            return

    gconf.__dict__.update(defaults.__dict__)
    gcnf.update_to(gconf.__dict__)
    gconf.__dict__.update(opts.__dict__)
    gconf.configinterface = gcnf

    delete = rconf.get('delete')
    if delete:
        logging.info('geo-replication delete')
        # Delete pid file, status file, socket file
        cleanup_paths = []
        if getattr(gconf, 'pid_file', None):
            cleanup_paths.append(gconf.pid_file)

        if getattr(gconf, 'state_file', None):
            cleanup_paths.append(gconf.state_file)

        if getattr(gconf, 'state_detail_file', None):
            cleanup_paths.append(gconf.state_detail_file)

        if getattr(gconf, 'state_socket_unencoded', None):
            cleanup_paths.append(gconf.state_socket_unencoded)

        cleanup_paths.append(rconf['config_file'][:-11] + "*")

        # Cleanup changelog working dirs
        if getattr(gconf, 'working_dir', None):
            try:
                shutil.rmtree(gconf.working_dir)
            except (IOError, OSError):
                if sys.exc_info()[1].errno == ENOENT:
                    pass
                else:
                    raise GsyncdError(
                        'Error while removing working dir: %s' %
                        gconf.working_dir)

        for path in cleanup_paths:
            # To delete temp files
            for f in glob.glob(path + "*"):
                _unlink(f)
        return

    if restricted and gconf.allow_network:
        ssh_conn = os.getenv('SSH_CONNECTION')
        if not ssh_conn:
            # legacy env var
            ssh_conn = os.getenv('SSH_CLIENT')
        if ssh_conn:
            allowed_networks = [IPNetwork(a)
                                for a in gconf.allow_network.split(',')]
            client_ip = IPAddress(ssh_conn.split()[0])
            allowed = False
            for nw in allowed_networks:
                if client_ip in nw:
                    allowed = True
                    break
            if not allowed:
                raise GsyncdError("client IP address is not allowed")

    ffd = rconf.get('feedback_fd')
    if ffd:
        fcntl.fcntl(ffd, fcntl.F_SETFD, fcntl.FD_CLOEXEC)

    # normalize loglevel
    lvl0 = gconf.log_level
    if isinstance(lvl0, str):
        lvl1 = lvl0.upper()
        lvl2 = logging.getLevelName(lvl1)
        # I have _never_ _ever_ seen such an utterly braindead
        # error condition
        if lvl2 == "Level " + lvl1:
            raise GsyncdError('cannot recognize log level "%s"' % lvl0)
        gconf.log_level = lvl2

    if not privileged() and gconf.log_file_mbr:
        gconf.log_file = gconf.log_file_mbr

    if checkpoint_change:
        try:
            GLogger._gsyncd_loginit(log_file=gconf.log_file, label='conf')
            if confdata.op == 'set':
                logging.info('checkpoint %s set' % confdata.val)
            elif confdata.op == 'del':
                logging.info('checkpoint info was reset')
        except IOError:
            if sys.exc_info()[1].errno == ENOENT:
                # directory of log path is not present,
                # which happens if we get here from
                # a peer-multiplexed "config-set checkpoint"
                # (as that directory is created only on the
                # original node)
                pass
            else:
                raise
        return

    create = rconf.get('create')
    if create:
        if getattr(gconf, 'state_file', None):
            set_monitor_status(gconf.state_file, create)
        return

    go_daemon = rconf['go_daemon']
    be_monitor = rconf.get('monitor')
    be_agent = rconf.get('agent')

    rscs, local, remote = makersc(args)

    status_get = rconf.get('status_get')
    if status_get:
        for brick in gconf.path:
            brick_status = GeorepStatus(gconf.state_file, brick)
            checkpoint_time = int(getattr(gconf, "checkpoint", "0"))
            brick_status.print_status(checkpoint_time=checkpoint_time)
        return

    if not be_monitor and isinstance(remote, resource.SSH) and \
       go_daemon == 'should':
        go_daemon = 'postconn'
        log_file = None
    else:
        log_file = gconf.log_file
    if be_monitor:
        label = 'monitor'
    elif be_agent:
        label = 'agent'
    elif remote:
        # master
        label = gconf.local_path
    else:
        label = 'slave'
    startup(go_daemon=go_daemon, log_file=log_file, label=label)
    resource.Popen.init_errhandler()

    if be_agent:
        os.setsid()
        logging.debug('rpc_fd: %s' % repr(gconf.rpc_fd))
        return agent(Changelog(), gconf.rpc_fd)

    if be_monitor:
        return monitor(*rscs)

    logging.info("syncing: %s" % " -> ".join(r.url for r in rscs))
    if remote:
        go_daemon = remote.connect_remote(go_daemon=go_daemon)
        if go_daemon:
            startup(go_daemon=go_daemon, log_file=gconf.log_file)
            # complete remote connection in child
            remote.connect_remote(go_daemon='done')
    local.connect()
    if ffd:
        os.close(ffd)
    local.service_loop(*[r for r in [remote] if r])


if __name__ == "__main__":
    main()



import sys
from uuid import UUID
from gluster.glupy import *

# Negative-lookup-caching example.  If a file wasn't there the last time we
# looked, it's probably still not there.  This translator keeps track of
# those failed lookups for us, and returns ENOENT without needing to pass the
# call any further for repeated requests.

# If we were doing this for real, we'd need separate caches for each xlator
# instance.  The easiest way to do this would be to have xlator.__init__
# "register" each instance in a module-global dict, with the key as the C
# translator address and the value as the xlator object itself.  For testing
# and teaching, it's sufficient just to have one cache.  The keys are parent
# GFIDs, and the entries are lists of names within that parent that we know
# don't exist.
cache = {}

# TBD: we need a better way of handling per-request data (frame->local in C).
dl.get_id.restype = c_long
dl.get_id.argtypes = [ POINTER(call_frame_t) ]

def uuid2str (gfid):
    return str(UUID(''.join(map("{0:02x}".format, gfid))))

class xlator (Translator):

    def __init__ (self, c_this):
        self.requests = {}
        Translator.__init__(self,c_this)

    def lookup_fop (self, frame, this, loc, xdata):
        pargfid = uuid2str(loc.contents.pargfid)
        print "lookup FOP: %s:%s" % (pargfid, loc.contents.name)
        # Check the cache.
        if cache.has_key(pargfid):
            if loc.contents.name in cache[pargfid]:
                print "short-circuiting for %s:%s" % (pargfid,
                    loc.contents.name)
                dl.unwind_lookup(frame,0,this,-1,2,None,None,None,None)
                return 0
        key = dl.get_id(frame)
        self.requests[key] = (pargfid, loc.contents.name[:])
        # TBD: get real child xl from init, pass it here
        dl.wind_lookup(frame,POINTER(xlator_t)(),loc,xdata)
        return 0

    def lookup_cbk (self, frame, cookie, this, op_ret, op_errno, inode, buf,
                    xdata, postparent):
        print "lookup CBK: %d (%d)" % (op_ret, op_errno)
        key = dl.get_id(frame)
        pargfid, name = self.requests[key]
        # Update the cache.
        if op_ret == 0:
            print "found %s, removing from cache" % name
            if cache.has_key(pargfid):
                cache[pargfid].discard(name)
        elif op_errno == 2:    # ENOENT
            print "failed to find %s, adding to cache" % name
            if cache.has_key(pargfid):
                cache[pargfid].add(name)
            else:
                cache[pargfid] = set([name])
        del self.requests[key]
        dl.unwind_lookup(frame,cookie,this,op_ret,op_errno,
                         inode,buf,xdata,postparent)
        return 0

    def create_fop (self, frame, this, loc, flags, mode, umask, fd, xdata):
        pargfid = uuid2str(loc.contents.pargfid)
        print "create FOP: %s:%s" % (pargfid, loc.contents.name)
        key = dl.get_id(frame)
        self.requests[key] = (pargfid, loc.contents.name[:])
        # TBD: get real child xl from init, pass it here
        dl.wind_create(frame,POINTER(xlator_t)(),loc,flags,mode,umask,fd,xdata)
        return 0

    def create_cbk (self, frame, cookie, this, op_ret, op_errno, fd, inode,
                    buf, preparent, postparent, xdata):
        print "create CBK: %d (%d)" % (op_ret, op_errno)
        key = dl.get_id(frame)
        pargfid, name = self.requests[key]
        # Update the cache.
        if op_ret == 0:
            print "created %s, removing from cache" % name
            if cache.has_key(pargfid):
                cache[pargfid].discard(name)
        del self.requests[key]
        dl.unwind_create(frame,cookie,this,op_ret,op_errno,fd,inode,buf,
                         preparent,postparent,xdata)
        return 0



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import re
import os
import sys
import stat
import time
import signal
import fcntl
import errno
import types
import struct
import socket
import logging
import tempfile
import threading
import subprocess
from errno import EEXIST, ENOENT, ENODATA, ENOTDIR, ELOOP
from errno import EISDIR, ENOTEMPTY, ESTALE, EINVAL
from select import error as SelectError
import shutil

from gconf import gconf
import repce
from repce import RepceServer, RepceClient
from master import gmaster_builder
import syncdutils
from syncdutils import GsyncdError, select, privileged, boolify, funcode
from syncdutils import umask, entry2pb, gauxpfx, errno_wrap, lstat
from syncdutils import NoPurgeTimeAvailable, PartialHistoryAvailable
from syncdutils import ChangelogException
from syncdutils import CHANGELOG_AGENT_CLIENT_VERSION
from gsyncdstatus import GeorepStatus


UrlRX = re.compile('\A(\w+)://([^ *?[]*)\Z')
HostRX = re.compile('[a-z\d](?:[a-z\d.-]*[a-z\d])?', re.I)
UserRX = re.compile("[\w!\#$%&'*+-\/=?^_`{|}~]+")


def sup(x, *a, **kw):
    """a rubyesque "super" for python ;)

    invoke caller method in parent class with given args.
    """
    return getattr(super(type(x), x),
                   sys._getframe(1).f_code.co_name)(*a, **kw)


def desugar(ustr):
    """transform sugared url strings to standard <scheme>://<urlbody> form

    parsing logic enforces the constraint that sugared forms should contatin
    a ':' or a '/', which ensures that sugared urls do not conflict with
    gluster volume names.
    """
    m = re.match('([^:]*):(.*)', ustr)
    if m:
        if not m.groups()[0]:
            return "gluster://localhost" + ustr
        elif '@' in m.groups()[0] or re.search('[:/]', m.groups()[1]):
            return "ssh://" + ustr
        else:
            return "gluster://" + ustr
    else:
        if ustr[0] != '/':
            raise GsyncdError("cannot resolve sugared url '%s'" % ustr)
        ap = os.path.normpath(ustr)
        if ap.startswith('//'):
            ap = ap[1:]
        return "file://" + ap


def gethostbyname(hnam):
    """gethostbyname wrapper"""
    try:
        return socket.gethostbyname(hnam)
    except socket.gaierror:
        ex = sys.exc_info()[1]
        raise GsyncdError("failed to resolve %s: %s" %
                          (hnam, ex.strerror))


def parse_url(ustr):
    """instantiate an url object by scheme-to-class dispatch

    The url classes taken into consideration are the ones in
    this module whose names are full-caps.
    """
    m = UrlRX.match(ustr)
    if not m:
        ustr = desugar(ustr)
    m = UrlRX.match(ustr)
    if not m:
        raise GsyncdError("malformed url")
    sch, path = m.groups()
    this = sys.modules[__name__]
    if not hasattr(this, sch.upper()):
        raise GsyncdError("unknown url scheme " + sch)
    return getattr(this, sch.upper())(path)


class _MetaXattr(object):

    """singleton class, a lazy wrapper around the
    libcxattr module

    libcxattr (a heavy import due to ctypes) is
    loaded only when when the single
    instance is tried to be used.

    This reduces runtime for those invocations
    which do not need filesystem manipulation
    (eg. for config, url parsing)
    """

    def __getattr__(self, meth):
        from libcxattr import Xattr as LXattr
        xmeth = [m for m in dir(LXattr) if m[0] != '_']
        if not meth in xmeth:
            return
        for m in xmeth:
            setattr(self, m, getattr(LXattr, m))
        return getattr(self, meth)


Xattr = _MetaXattr()


class Popen(subprocess.Popen):

    """customized subclass of subprocess.Popen with a ring
    buffer for children error output"""

    @classmethod
    def init_errhandler(cls):
        """start the thread which handles children's error output"""
        cls.errstore = {}

        def tailer():
            while True:
                errstore = cls.errstore.copy()
                try:
                    poe, _, _ = select(
                        [po.stderr for po in errstore], [], [], 1)
                except (ValueError, SelectError):
                    continue
                for po in errstore:
                    if po.stderr not in poe:
                        continue
                    po.lock.acquire()
                    try:
                        if po.on_death_row:
                            continue
                        la = errstore[po]
                        try:
                            fd = po.stderr.fileno()
                        except ValueError:  # file is already closed
                            continue
                        l = os.read(fd, 1024)
                        if not l:
                            continue
                        tots = len(l)
                        for lx in la:
                            tots += len(lx)
                        while tots > 1 << 20 and la:
                            tots -= len(la.pop(0))
                        la.append(l)
                    finally:
                        po.lock.release()
        t = syncdutils.Thread(target=tailer)
        t.start()
        cls.errhandler = t

    @classmethod
    def fork(cls):
        """fork wrapper that restarts errhandler thread in child"""
        pid = os.fork()
        if not pid:
            cls.init_errhandler()
        return pid

    def __init__(self, args, *a, **kw):
        """customizations for subprocess.Popen instantiation

        - 'close_fds' is taken to be the default
        - if child's stderr is chosen to be managed,
          register it with the error handler thread
        """
        self.args = args
        if 'close_fds' not in kw:
            kw['close_fds'] = True
        self.lock = threading.Lock()
        self.on_death_row = False
        try:
            sup(self, args, *a, **kw)
        except:
            ex = sys.exc_info()[1]
            if not isinstance(ex, OSError):
                raise
            raise GsyncdError("""execution of "%s" failed with %s (%s)""" %
                              (args[0], errno.errorcode[ex.errno],
                               os.strerror(ex.errno)))
        if kw.get('stderr') == subprocess.PIPE:
            assert(getattr(self, 'errhandler', None))
            self.errstore[self] = []

    def errlog(self):
        """make a log about child's failure event"""
        filling = ""
        if self.elines:
            filling = ", saying:"
        logging.error("""command "%s" returned with %s%s""" %
                      (" ".join(self.args), repr(self.returncode), filling))
        lp = ''

        def logerr(l):
            logging.error(self.args[0] + "> " + l)
        for l in self.elines:
            ls = l.split('\n')
            ls[0] = lp + ls[0]
            lp = ls.pop()
            for ll in ls:
                logerr(ll)
        if lp:
            logerr(lp)

    def errfail(self):
        """fail nicely if child did not terminate with success"""
        self.errlog()
        syncdutils.finalize(exval=1)

    def terminate_geterr(self, fail_on_err=True):
        """kill child, finalize stderr harvesting (unregister
        from errhandler, set up .elines), fail on error if
        asked for
        """
        self.lock.acquire()
        try:
            self.on_death_row = True
        finally:
            self.lock.release()
        elines = self.errstore.pop(self)
        if self.poll() is None:
            self.terminate()
            if self.poll() is None:
                time.sleep(0.1)
                self.kill()
                self.wait()
        while True:
            if not select([self.stderr], [], [], 0.1)[0]:
                break
            b = os.read(self.stderr.fileno(), 1024)
            if b:
                elines.append(b)
            else:
                break
        self.stderr.close()
        self.elines = elines
        if fail_on_err and self.returncode != 0:
            self.errfail()


class Server(object):

    """singleton implemening those filesystem access primitives
       which are needed for geo-replication functionality

    (Singleton in the sense it's a class which has only static
    and classmethods and is used directly, without instantiation.)
    """

    GX_NSPACE_PFX = (privileged() and "trusted" or "system")
    GX_NSPACE = GX_NSPACE_PFX + ".glusterfs"
    NTV_FMTSTR = "!" + "B" * 19 + "II"
    FRGN_XTRA_FMT = "I"
    FRGN_FMTSTR = NTV_FMTSTR + FRGN_XTRA_FMT
    GX_GFID_CANONICAL_LEN = 37  # canonical gfid len + '\0'

    # for backend gfid fetch, do not use GX_NSPACE_PFX
    GFID_XATTR = 'trusted.gfid'
    GFID_FMTSTR = "!" + "B" * 16

    local_path = ''

    @classmethod
    def _fmt_mknod(cls, l):
        return "!II%dsI%dsIII" % (cls.GX_GFID_CANONICAL_LEN, l + 1)

    @classmethod
    def _fmt_mkdir(cls, l):
        return "!II%dsI%dsII" % (cls.GX_GFID_CANONICAL_LEN, l + 1)

    @classmethod
    def _fmt_symlink(cls, l1, l2):
        return "!II%dsI%ds%ds" % (cls.GX_GFID_CANONICAL_LEN, l1 + 1, l2 + 1)

    def _pathguard(f):
        """decorator method that checks
        the path argument of the decorated
        functions to make sure it does not
        point out of the managed tree
        """

        fc = funcode(f)
        pi = list(fc.co_varnames).index('path')

        def ff(*a):
            path = a[pi]
            ps = path.split('/')
            if path[0] == '/' or '..' in ps:
                raise ValueError('unsafe path')
            a = list(a)
            a[pi] = os.path.join(a[0].local_path, path)
            return f(*a)
        return ff

    @classmethod
    @_pathguard
    def entries(cls, path):
        """directory entries in an array"""
        # prevent symlinks being followed
        if not stat.S_ISDIR(os.lstat(path).st_mode):
            raise OSError(ENOTDIR, os.strerror(ENOTDIR))
        return os.listdir(path)

    @classmethod
    @_pathguard
    def lstat(cls, path):
        try:
            return os.lstat(path)
        except (IOError, OSError):
            ex = sys.exc_info()[1]
            if ex.errno == ENOENT:
                return ex.errno
            else:
                raise

    @classmethod
    @_pathguard
    def linkto_check(cls, path):
        try:
            return not (
                Xattr.lgetxattr_buf(path,
                                    'trusted.glusterfs.dht.linkto') == '')
        except (IOError, OSError):
            ex = sys.exc_info()[1]
            if ex.errno in (ENOENT, ENODATA):
                return False
            else:
                raise

    @classmethod
    @_pathguard
    def gfid(cls, path):
        try:
            buf = Xattr.lgetxattr(path, cls.GFID_XATTR, 16)
            m = re.match('(.{8})(.{4})(.{4})(.{4})(.{12})', "".join(
                ['%02x' % x for x in struct.unpack(cls.GFID_FMTSTR, buf)]))
            return '-'.join(m.groups())
        except (IOError, OSError):
            ex = sys.exc_info()[1]
            if ex.errno == ENOENT:
                return ex.errno
            else:
                raise

    @classmethod
    def gfid_mnt(cls, gfidpath):
        return errno_wrap(Xattr.lgetxattr,
                          [gfidpath, 'glusterfs.gfid.string',
                           cls.GX_GFID_CANONICAL_LEN], [ENOENT])

    @classmethod
    @_pathguard
    def purge(cls, path, entries=None):
        """force-delete subtrees

        If @entries is not specified, delete
        the whole subtree under @path (including
        @path).

        Otherwise, @entries should be a
        a sequence of children of @path, and
        the effect is identical with a joint
        @entries-less purge on them, ie.

        for e in entries:
            cls.purge(os.path.join(path, e))
        """
        me_also = entries is None
        if not entries:
            try:
                # if it's a symlink, prevent
                # following it
                try:
                    os.unlink(path)
                    return
                except OSError:
                    ex = sys.exc_info()[1]
                    if ex.errno == EISDIR:
                        entries = os.listdir(path)
                    else:
                        raise
            except OSError:
                ex = sys.exc_info()[1]
                if ex.errno in (ENOTDIR, ENOENT, ELOOP):
                    try:
                        os.unlink(path)
                        return
                    except OSError:
                        ex = sys.exc_info()[1]
                        if ex.errno == ENOENT:
                            return
                        raise
                else:
                    raise
        for e in entries:
            cls.purge(os.path.join(path, e))
        if me_also:
            os.rmdir(path)

    @classmethod
    @_pathguard
    def _create(cls, path, ctor):
        """path creation backend routine"""
        try:
            ctor(path)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno == EEXIST:
                cls.purge(path)
                return ctor(path)
            raise

    @classmethod
    @_pathguard
    def mkdir(cls, path):
        cls._create(path, os.mkdir)

    @classmethod
    @_pathguard
    def symlink(cls, lnk, path):
        cls._create(path, lambda p: os.symlink(lnk, p))

    @classmethod
    @_pathguard
    def xtime(cls, path, uuid):
        """query xtime extended attribute

        Return xtime of @path for @uuid as a pair of integers.
        "Normal" errors due to non-existent @path or extended attribute
        are tolerated and errno is returned in such a case.
        """

        try:
            val = Xattr.lgetxattr(path,
                                  '.'.join([cls.GX_NSPACE, uuid, 'xtime']),
                                  8)
            return struct.unpack('!II', val)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno in (ENOENT, ENODATA, ENOTDIR):
                return ex.errno
            else:
                raise

    @classmethod
    @_pathguard
    def stime_mnt(cls, path, uuid):
        """query xtime extended attribute

        Return xtime of @path for @uuid as a pair of integers.
        "Normal" errors due to non-existent @path or extended attribute
        are tolerated and errno is returned in such a case.
        """

        try:
            val = Xattr.lgetxattr(path,
                                  '.'.join([cls.GX_NSPACE, uuid, 'stime']),
                                  8)
            return struct.unpack('!II', val)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno in (ENOENT, ENODATA, ENOTDIR):
                return ex.errno
            else:
                raise

    @classmethod
    @_pathguard
    def stime(cls, path, uuid):
        """query xtime extended attribute

        Return xtime of @path for @uuid as a pair of integers.
        "Normal" errors due to non-existent @path or extended attribute
        are tolerated and errno is returned in such a case.
        """

        try:
            val = Xattr.lgetxattr(path,
                                  '.'.join([cls.GX_NSPACE, uuid, 'stime']),
                                  8)
            return struct.unpack('!II', val)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno in (ENOENT, ENODATA, ENOTDIR):
                return ex.errno
            else:
                raise

    @classmethod
    def node_uuid(cls, path='.'):
        try:
            uuid_l = Xattr.lgetxattr_buf(
                path, '.'.join([cls.GX_NSPACE, 'node-uuid']))
            return uuid_l[:-1].split(' ')
        except OSError:
            raise

    @classmethod
    @_pathguard
    def set_stime(cls, path, uuid, mark):
        """set @mark as stime for @uuid on @path"""
        Xattr.lsetxattr(
            path, '.'.join([cls.GX_NSPACE, uuid, 'stime']),
            struct.pack('!II', *mark))

    @classmethod
    @_pathguard
    def set_xtime(cls, path, uuid, mark):
        """set @mark as xtime for @uuid on @path"""
        Xattr.lsetxattr(
            path, '.'.join([cls.GX_NSPACE, uuid, 'xtime']),
            struct.pack('!II', *mark))

    @classmethod
    @_pathguard
    def set_xtime_remote(cls, path, uuid, mark):
        """
        set @mark as xtime for @uuid on @path
        the difference b/w this and set_xtime() being
        set_xtime() being overloaded to set the xtime
        on the brick (this method sets xtime on the
        remote slave)
        """
        Xattr.lsetxattr(
            path, '.'.join([cls.GX_NSPACE, uuid, 'xtime']),
            struct.pack('!II', *mark))

    @classmethod
    def entry_ops(cls, entries):
        pfx = gauxpfx()
        logging.debug('entries: %s' % repr(entries))
        # regular file

        def entry_pack_reg(gf, bn, mo, uid, gid):
            blen = len(bn)
            return struct.pack(cls._fmt_mknod(blen),
                               uid, gid, gf, mo, bn,
                               stat.S_IMODE(mo), 0, umask())

        def entry_pack_reg_stat(gf, bn, st):
            blen = len(bn)
            mo = st['mode']
            return struct.pack(cls._fmt_mknod(blen),
                               st['uid'], st['gid'],
                               gf, mo, bn,
                               stat.S_IMODE(mo), 0, umask())
        # mkdir

        def entry_pack_mkdir(gf, bn, mo, uid, gid):
            blen = len(bn)
            return struct.pack(cls._fmt_mkdir(blen),
                               uid, gid, gf, mo, bn,
                               stat.S_IMODE(mo), umask())
        # symlink

        def entry_pack_symlink(gf, bn, lnk, st):
            blen = len(bn)
            llen = len(lnk)
            return struct.pack(cls._fmt_symlink(blen, llen),
                               st['uid'], st['gid'],
                               gf, st['mode'], bn, lnk)

        def entry_purge(entry, gfid):
            # This is an extremely racy code and needs to be fixed ASAP.
            # The GFID check here is to be sure that the pargfid/bname
            # to be purged is the GFID gotten from the changelog.
            # (a stat(changelog_gfid) would also be valid here)
            # The race here is between the GFID check and the purge.
            if not matching_disk_gfid(gfid, entry):
                return

            er = errno_wrap(os.unlink, [entry], [ENOENT, ESTALE, EISDIR])
            if isinstance(er, int):
                if er == EISDIR:
                    er = errno_wrap(os.rmdir, [entry], [ENOENT, ESTALE,
                                                        ENOTEMPTY])
                    if er == ENOTEMPTY:
                        return er

        def collect_failure(e, cmd_ret):
            # We do this for failing fops on Slave
            # Master should be logging this
            if cmd_ret is None:
                return

            if cmd_ret == EEXIST:
                disk_gfid = cls.gfid_mnt(e['entry'])
                if isinstance(disk_gfid, basestring):
                    if e['gfid'] != disk_gfid:
                        failures.append((e, cmd_ret, disk_gfid))
            else:
                failures.append((e, cmd_ret))

        failures = []

        def matching_disk_gfid(gfid, entry):
            disk_gfid = cls.gfid_mnt(entry)
            if isinstance(disk_gfid, int):
                return False

            if not gfid == disk_gfid:
                return False

            return True

        def recursive_rmdir(gfid, entry, path):
            """disk_gfid check added for original path for which
            recursive_delete is called. This disk gfid check executed
            before every Unlink/Rmdir. If disk gfid is not matching
            with GFID from Changelog, that means other worker
            deleted the directory. Even if the subdir/file present,
            it belongs to different parent. Exit without performing
            further deletes.
            """
            if not matching_disk_gfid(gfid, entry):
                return

            names = []
            names = errno_wrap(os.listdir, [path], [ENOENT])
            if isinstance(names, int):
                return

            for name in names:
                fullname = os.path.join(path, name)
                if not matching_disk_gfid(gfid, entry):
                    return
                er = errno_wrap(os.remove, [fullname], [ENOENT, ESTALE,
                                                        EISDIR])

                if er == EISDIR:
                    recursive_rmdir(gfid, entry, fullname)

            if not matching_disk_gfid(gfid, entry):
                return

            errno_wrap(os.rmdir, [path], [ENOENT, ESTALE])

        for e in entries:
            blob = None
            op = e['op']
            gfid = e['gfid']
            entry = e['entry']
            (pg, bname) = entry2pb(entry)
            if op in ['RMDIR', 'UNLINK']:
                # Try once, if rmdir failed with ENOTEMPTY
                # then delete recursively.
                er = entry_purge(entry, gfid)
                if isinstance(er, int):
                    if er == ENOTEMPTY and op == 'RMDIR':
                        # Retry if ENOTEMPTY, ESTALE
                        er1 = errno_wrap(recursive_rmdir,
                                         [gfid, entry,
                                          os.path.join(pg, bname)],
                                         [], [ENOTEMPTY, ESTALE, ENODATA])
                        if not isinstance(er1, int):
                            logging.debug("Removed %s => %s/%s recursively" %
                                          (gfid, pg, bname))
                        else:
                            logging.warn("Recursive remove %s => %s/%s"
                                         "failed: %s" % (gfid, pg, bname,
                                                         os.strerror(er1)))
                    else:
                        logging.warn("Failed to remove %s => %s/%s. %s" %
                                     (gfid, pg, bname, os.strerror(er)))
            elif op in ['CREATE', 'MKNOD']:
                blob = entry_pack_reg(
                    gfid, bname, e['mode'], e['uid'], e['gid'])
            elif op == 'MKDIR':
                blob = entry_pack_mkdir(
                    gfid, bname, e['mode'], e['uid'], e['gid'])
            elif op == 'LINK':
                slink = os.path.join(pfx, gfid)
                st = lstat(slink)
                if isinstance(st, int):
                    (pg, bname) = entry2pb(entry)
                    blob = entry_pack_reg_stat(gfid, bname, e['stat'])
                else:
                    cmd_ret = errno_wrap(os.link,
                                         [slink, entry],
                                         [ENOENT, EEXIST])
                    collect_failure(e, cmd_ret)
            elif op == 'SYMLINK':
                blob = entry_pack_symlink(gfid, bname, e['link'], e['stat'])
            elif op == 'RENAME':
                en = e['entry1']
                st = lstat(entry)
                if isinstance(st, int):
                    if e['stat'] and not stat.S_ISDIR(e['stat']['mode']):
                        (pg, bname) = entry2pb(en)
                        blob = entry_pack_reg_stat(gfid, bname, e['stat'])
                else:
                    cmd_ret = errno_wrap(os.rename,
                                         [entry, en],
                                         [ENOENT, EEXIST])
                    collect_failure(e, cmd_ret)
            if blob:
                cmd_ret = errno_wrap(Xattr.lsetxattr,
                                     [pg, 'glusterfs.gfid.newfile', blob],
                                     [EEXIST, ENOENT],
                                     [ESTALE, EINVAL])
                collect_failure(e, cmd_ret)
        return failures

    @classmethod
    def meta_ops(cls, meta_entries):
        logging.debug('Meta-entries: %s' % repr(meta_entries))
        failures = []
        for e in meta_entries:
            mode = e['stat']['mode']
            uid = e['stat']['uid']
            gid = e['stat']['gid']
            atime = e['stat']['atime']
            mtime = e['stat']['mtime']
            go = e['go']
            cmd_ret = errno_wrap(os.chmod, [go, mode],
                                 [ENOENT], [ESTALE, EINVAL])
            # This is a fail fast mechanism
            # We do this for failing fops on Slave
            # Master should be logging this
            if isinstance(cmd_ret, int):
                failures.append((e, cmd_ret))
                continue
            errno_wrap(os.chown, [go, uid, gid], [ENOENT], [ESTALE, EINVAL])
            errno_wrap(os.utime, [go, (atime, mtime)],
                       [ENOENT], [ESTALE, EINVAL])
        return failures

    @classmethod
    @_pathguard
    def setattr(cls, path, adct):
        """set file attributes

        @adct is a dict, where 'own', 'mode' and 'times'
        keys are looked for and values used to perform
        chown, chmod or utimes on @path.
        """
        own = adct.get('own')
        if own:
            os.lchown(path, *own)
        mode = adct.get('mode')
        if mode:
            os.chmod(path, stat.S_IMODE(mode))
        times = adct.get('times')
        if times:
            os.utime(path, times)

    @staticmethod
    def pid():
        return os.getpid()

    last_keep_alive = 0

    @classmethod
    def keep_alive(cls, dct):
        """process keepalive messages.

        Return keep-alive counter (number of received keep-alive
        messages).

        Now the "keep-alive" message can also have a payload which is
        used to set a foreign volume-mark on the underlying file system.
        """
        if dct:
            key = '.'.join([cls.GX_NSPACE, 'volume-mark', dct['uuid']])
            val = struct.pack(cls.FRGN_FMTSTR,
                              *(dct['version'] +
                                tuple(int(x, 16)
                                      for x in re.findall('(?:[\da-f]){2}',
                                                          dct['uuid'])) +
                                (dct['retval'],) + dct['volume_mark'][0:2] + (
                                    dct['timeout'],)))
            Xattr.lsetxattr('.', key, val)
        cls.last_keep_alive += 1
        return cls.last_keep_alive

    @staticmethod
    def version():
        """version used in handshake"""
        return 1.0


class SlaveLocal(object):

    """mix-in class to implement some factes of a slave server

    ("mix-in" is sort of like "abstract class", ie. it's not
    instantiated just included in the ancesty DAG. I use "mix-in"
    to indicate that it's not used as an abstract base class,
    rather just taken in to implement additional functionality
    on the basis of the assumed availability of certain interfaces.)
    """

    def can_connect_to(self, remote):
        """determine our position in the connectibility matrix"""
        return not remote

    def service_loop(self):
        """start a RePCe server serving self's server

        stop servicing if a timeout is configured and got no
        keep-alime in that inteval
        """

        if boolify(gconf.use_rsync_xattrs) and not privileged():
            raise GsyncdError(
                "using rsync for extended attributes is not supported")

        repce = RepceServer(
            self.server, sys.stdin, sys.stdout, int(gconf.sync_jobs))
        t = syncdutils.Thread(target=lambda: (repce.service_loop(),
                                              syncdutils.finalize()))
        t.start()
        logging.info("slave listening")
        if gconf.timeout and int(gconf.timeout) > 0:
            while True:
                lp = self.server.last_keep_alive
                time.sleep(int(gconf.timeout))
                if lp == self.server.last_keep_alive:
                    logging.info(
                        "connection inactive for %d seconds, stopping" %
                        int(gconf.timeout))
                    break
        else:
            select((), (), ())


class SlaveRemote(object):

    """mix-in class to implement an interface to a remote slave"""

    def connect_remote(self, rargs=[], **opts):
        """connects to a remote slave

        Invoke an auxiliary utility (slave gsyncd, possibly wrapped)
        which sets up the connection and set up a RePCe client to
        communicate throuh its stdio.
        """
        slave = opts.get('slave', self.url)
        extra_opts = []
        so = getattr(gconf, 'session_owner', None)
        if so:
            extra_opts += ['--session-owner', so]
        if boolify(gconf.use_rsync_xattrs):
            extra_opts.append('--use-rsync-xattrs')
        po = Popen(rargs + gconf.remote_gsyncd.split() + extra_opts +
                   ['-N', '--listen', '--timeout', str(gconf.timeout),
                    slave],
                   stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                   stderr=subprocess.PIPE)
        gconf.transport = po
        return self.start_fd_client(po.stdout, po.stdin, **opts)

    def start_fd_client(self, i, o, **opts):
        """set up RePCe client, handshake with server

        It's cut out as a separate method to let
        subclasses hook into client startup
        """
        self.server = RepceClient(i, o)
        rv = self.server.__version__()
        exrv = {'proto': repce.repce_version, 'object': Server.version()}
        da0 = (rv, exrv)
        da1 = ({}, {})
        for i in range(2):
            for k, v in da0[i].iteritems():
                da1[i][k] = int(v)
        if da1[0] != da1[1]:
            raise GsyncdError(
                "RePCe major version mismatch: local %s, remote %s" %
                (exrv, rv))

    def rsync(self, files, *args):
        """invoke rsync"""
        if not files:
            raise GsyncdError("no files to sync")
        logging.debug("files: " + ", ".join(files))
        argv = gconf.rsync_command.split() + \
            ['-aR0', '--inplace', '--files-from=-', '--super',
             '--stats', '--numeric-ids', '--no-implied-dirs'] + \
            gconf.rsync_options.split() + \
            (boolify(gconf.sync_xattrs) and ['--xattrs'] or []) + \
            (boolify(gconf.sync_acls) and ['--acls'] or []) + \
            ['.'] + list(args)

        if gconf.log_rsync_performance:
            # use stdout=PIPE only when log_rsync_performance enabled
            # Else rsync will write to stdout and nobody is their
            # to consume. If PIPE is full rsync hangs.
            po = Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE)
        else:
            po = Popen(argv, stdin=subprocess.PIPE, stderr=subprocess.PIPE)

        for f in files:
            po.stdin.write(f)
            po.stdin.write('\0')

        po.stdin.close()

        if gconf.log_rsync_performance:
            out = po.stdout.read()
            rsync_msg = []
            for line in out.split("\n"):
                if line.startswith("Number of files:") or \
                   line.startswith("Number of regular files transferred:") or \
                   line.startswith("Total file size:") or \
                   line.startswith("Total transferred file size:") or \
                   line.startswith("Literal data:") or \
                   line.startswith("Matched data:") or \
                   line.startswith("Total bytes sent:") or \
                   line.startswith("Total bytes received:") or \
                   line.startswith("sent "):
                    rsync_msg.append(line)
            logging.info("rsync performance: %s" % ", ".join(rsync_msg))
        po.wait()
        po.terminate_geterr(fail_on_err=False)

        return po

    def tarssh(self, files, slaveurl):
        """invoke tar+ssh
        -z (compress) can be use if needed, but omitting it now
        as it results in weird error (tar+ssh errors out (errcode: 2)
        """
        if not files:
            raise GsyncdError("no files to sync")
        logging.debug("files: " + ", ".join(files))
        (host, rdir) = slaveurl.split(':')
        tar_cmd = ["tar"] + \
            ["-cf", "-", "--files-from", "-"]
        ssh_cmd = gconf.ssh_command_tar.split() + \
            [host, "tar"] + \
            ["--overwrite", "-xf", "-", "-C", rdir]
        p0 = Popen(tar_cmd, stdout=subprocess.PIPE,
                   stdin=subprocess.PIPE, stderr=subprocess.PIPE)
        p1 = Popen(ssh_cmd, stdin=p0.stdout, stderr=subprocess.PIPE)
        for f in files:
            p0.stdin.write(f)
            p0.stdin.write('\n')
        p0.stdin.close()

        # wait() for tar to terminate, collecting any errors, further
        # waiting for transfer to complete
        p0.wait()
        p0.terminate_geterr(fail_on_err=False)

        p1.wait()
        p1.terminate_geterr(fail_on_err=False)

        return p1


class AbstractUrl(object):

    """abstract base class for url scheme classes"""

    def __init__(self, path, pattern):
        m = re.search(pattern, path)
        if not m:
            raise GsyncdError("malformed path")
        self.path = path
        return m.groups()

    @property
    def scheme(self):
        return type(self).__name__.lower()

    def canonical_path(self):
        return self.path

    def get_url(self, canonical=False, escaped=False):
        """format self's url in various styles"""
        if canonical:
            pa = self.canonical_path()
        else:
            pa = self.path
        u = "://".join((self.scheme, pa))
        if escaped:
            u = syncdutils.escape(u)
        return u

    @property
    def url(self):
        return self.get_url()


class FILE(AbstractUrl, SlaveLocal, SlaveRemote):

    """scheme class for file:// urls

    can be used to represent a file slave server
    on slave side, or interface to a remote file
    file server on master side
    """

    class FILEServer(Server):

        """included server flavor"""
        pass

    server = FILEServer

    def __init__(self, path):
        sup(self, path, '^/')

    def connect(self):
        """inhibit the resource beyond"""
        os.chdir(self.path)

    def rsync(self, files):
        return sup(self, files, self.path)


class GLUSTER(AbstractUrl, SlaveLocal, SlaveRemote):

    """scheme class for gluster:// urls

    can be used to represent a gluster slave server
    on slave side, or interface to a remote gluster
    slave on master side, or to represent master
    (slave-ish features come from the mixins, master
    functionality is outsourced to GMaster from master)
    """

    class GLUSTERServer(Server):

        "server enhancements for a glusterfs backend"""

        @classmethod
        def _attr_unpack_dict(cls, xattr, extra_fields=''):
            """generic volume mark fetching/parsing backed"""
            fmt_string = cls.NTV_FMTSTR + extra_fields
            buf = Xattr.lgetxattr('.', xattr, struct.calcsize(fmt_string))
            vm = struct.unpack(fmt_string, buf)
            m = re.match(
                '(.{8})(.{4})(.{4})(.{4})(.{12})',
                "".join(['%02x' % x for x in vm[2:18]]))
            uuid = '-'.join(m.groups())
            volinfo = {'version': vm[0:2],
                       'uuid': uuid,
                       'retval': vm[18],
                       'volume_mark': vm[19:21],
                       }
            if extra_fields:
                return volinfo, vm[-len(extra_fields):]
            else:
                return volinfo

        @classmethod
        def foreign_volume_infos(cls):
            """return list of valid (not expired) foreign volume marks"""
            dict_list = []
            xattr_list = Xattr.llistxattr_buf('.')
            for ele in xattr_list:
                if ele.find('.'.join([cls.GX_NSPACE, 'volume-mark', ''])) == 0:
                    d, x = cls._attr_unpack_dict(ele, cls.FRGN_XTRA_FMT)
                    now = int(time.time())
                    if x[0] > now:
                        logging.debug("volinfo[%s] expires: %d "
                                      "(%d sec later)" %
                                      (d['uuid'], x[0], x[0] - now))
                        d['timeout'] = x[0]
                        dict_list.append(d)
                    else:
                        try:
                            Xattr.lremovexattr('.', ele)
                        except OSError:
                            pass
            return dict_list

        @classmethod
        def native_volume_info(cls):
            """get the native volume mark of the underlying gluster volume"""
            try:
                return cls._attr_unpack_dict('.'.join([cls.GX_NSPACE,
                                                       'volume-mark']))
            except OSError:
                ex = sys.exc_info()[1]
                if ex.errno != ENODATA:
                    raise

    server = GLUSTERServer

    def __init__(self, path):
        self.host, self.volume = sup(self, path, '^(%s):(.+)' % HostRX.pattern)

    def canonical_path(self):
        return ':'.join([gethostbyname(self.host), self.volume])

    def can_connect_to(self, remote):
        """determine our position in the connectibility matrix"""
        return not remote or \
            (isinstance(remote, SSH) and isinstance(remote.inner_rsc, GLUSTER))

    class Mounter(object):

        """Abstract base class for mounter backends"""

        def __init__(self, params):
            self.params = params
            self.mntpt = None

        @classmethod
        def get_glusterprog(cls):
            return os.path.join(gconf.gluster_command_dir, cls.glusterprog)

        def umount_l(self, d):
            """perform lazy umount"""
            po = Popen(self.make_umount_argv(d), stderr=subprocess.PIPE)
            po.wait()
            return po

        @classmethod
        def make_umount_argv(cls, d):
            raise NotImplementedError

        def make_mount_argv(self, *a):
            raise NotImplementedError

        def cleanup_mntpt(self, *a):
            pass

        def handle_mounter(self, po):
            po.wait()

        def inhibit(self, *a):
            """inhibit a gluster filesystem

            Mount glusterfs over a temporary mountpoint,
            change into the mount, and lazy unmount the
            filesystem.
            """

            mpi, mpo = os.pipe()
            mh = Popen.fork()
            if mh:
                os.close(mpi)
                fcntl.fcntl(mpo, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
                d = None
                margv = self.make_mount_argv(*a)
                if self.mntpt:
                    # mntpt is determined pre-mount
                    d = self.mntpt
                    os.write(mpo, d + '\0')
                po = Popen(margv, **self.mountkw)
                self.handle_mounter(po)
                po.terminate_geterr()
                logging.debug('auxiliary glusterfs mount in place')
                if not d:
                    # mntpt is determined during mount
                    d = self.mntpt
                    os.write(mpo, d + '\0')
                os.write(mpo, 'M')
                t = syncdutils.Thread(target=lambda: os.chdir(d))
                t.start()
                tlim = gconf.starttime + int(gconf.connection_timeout)
                while True:
                    if not t.isAlive():
                        break
                    if time.time() >= tlim:
                        syncdutils.finalize(exval=1)
                    time.sleep(1)
                os.close(mpo)
                _, rv = syncdutils.waitpid(mh, 0)
                if rv:
                    rv = (os.WIFEXITED(rv) and os.WEXITSTATUS(rv) or 0) - \
                         (os.WIFSIGNALED(rv) and os.WTERMSIG(rv) or 0)
                    logging.warn('stale mount possibly left behind on ' + d)
                    raise GsyncdError("cleaning up temp mountpoint %s "
                                      "failed with status %d" %
                                      (d, rv))
            else:
                rv = 0
                try:
                    os.setsid()
                    os.close(mpo)
                    mntdata = ''
                    while True:
                        c = os.read(mpi, 1)
                        if not c:
                            break
                        mntdata += c
                    if mntdata:
                        mounted = False
                        if mntdata[-1] == 'M':
                            mntdata = mntdata[:-1]
                            assert(mntdata)
                            mounted = True
                        assert(mntdata[-1] == '\0')
                        mntpt = mntdata[:-1]
                        assert(mntpt)
                        if mounted:
                            po = self.umount_l(mntpt)
                            po.terminate_geterr(fail_on_err=False)
                            if po.returncode != 0:
                                po.errlog()
                                rv = po.returncode
                        self.cleanup_mntpt(mntpt)
                except:
                    logging.exception('mount cleanup failure:')
                    rv = 200
                os._exit(rv)
            logging.debug('auxiliary glusterfs mount prepared')

    class DirectMounter(Mounter):

        """mounter backend which calls mount(8), umount(8) directly"""

        mountkw = {'stderr': subprocess.PIPE}
        glusterprog = 'glusterfs'

        @staticmethod
        def make_umount_argv(d):
            return ['umount', '-l', d]

        def make_mount_argv(self):
            self.mntpt = tempfile.mkdtemp(prefix='gsyncd-aux-mount-')
            return [self.get_glusterprog()] + \
                ['--' + p for p in self.params] + [self.mntpt]

        def cleanup_mntpt(self, mntpt=None):
            if not mntpt:
                mntpt = self.mntpt
            os.rmdir(mntpt)

    class MountbrokerMounter(Mounter):

        """mounter backend using the mountbroker gluster service"""

        mountkw = {'stderr': subprocess.PIPE, 'stdout': subprocess.PIPE}
        glusterprog = 'gluster'

        @classmethod
        def make_cli_argv(cls):
            return [cls.get_glusterprog()] + ['--remote-host=localhost'] + \
                gconf.gluster_cli_options.split() + ['system::']

        @classmethod
        def make_umount_argv(cls, d):
            return cls.make_cli_argv() + ['umount', d, 'lazy']

        def make_mount_argv(self, label):
            return self.make_cli_argv() + \
                ['mount', label, 'user-map-root=' +
                    syncdutils.getusername()] + self.params

        def handle_mounter(self, po):
            self.mntpt = po.stdout.readline()[:-1]
            po.stdout.close()
            sup(self, po)
            if po.returncode != 0:
                # if cli terminated with error due to being
                # refused by glusterd, what it put
                # out on stdout is a diagnostic message
                logging.error('glusterd answered: %s' % self.mntpt)

    def connect(self):
        """inhibit the resource beyond

        Choose mounting backend (direct or mountbroker),
        set up glusterfs parameters and perform the mount
        with given backend
        """

        label = getattr(gconf, 'mountbroker', None)
        if not label and not privileged():
            label = syncdutils.getusername()
        mounter = label and self.MountbrokerMounter or self.DirectMounter
        params = gconf.gluster_params.split() + \
            (gconf.gluster_log_level and ['log-level=' +
                                          gconf.gluster_log_level] or []) + \
            ['log-file=' + gconf.gluster_log_file, 'volfile-server=' +
             self.host, 'volfile-id=' + self.volume, 'client-pid=-1']
        mounter(params).inhibit(*[l for l in [label] if l])

    def connect_remote(self, *a, **kw):
        sup(self, *a, **kw)
        self.slavedir = "/proc/%d/cwd" % self.server.pid()

    def gmaster_instantiate_tuple(self, slave):
        """return a tuple of the 'one shot' and the 'main crawl'
        class instance"""
        return (gmaster_builder('xsync')(self, slave),
                gmaster_builder()(self, slave),
                gmaster_builder('changeloghistory')(self, slave))

    def service_loop(self, *args):
        """enter service loop

        - if slave given, instantiate GMaster and
          pass control to that instance, which implements
          master behavior
        - else do that's what's inherited
        """
        if args:
            slave = args[0]
            if gconf.local_path:
                class brickserver(FILE.FILEServer):
                    local_path = gconf.local_path
                    aggregated = self.server

                    @classmethod
                    def entries(cls, path):
                        e = super(brickserver, cls).entries(path)
                        # on the brick don't mess with /.glusterfs
                        if path == '.':
                            try:
                                e.remove('.glusterfs')
                                e.remove('.trashcan')
                            except ValueError:
                                pass
                        return e

                    @classmethod
                    def lstat(cls, e):
                        """ path based backend stat """
                        return super(brickserver, cls).lstat(e)

                    @classmethod
                    def gfid(cls, e):
                        """ path based backend gfid fetch """
                        return super(brickserver, cls).gfid(e)

                    @classmethod
                    def linkto_check(cls, e):
                        return super(brickserver, cls).linkto_check(e)
                if gconf.slave_id:
                    # define {,set_}xtime in slave, thus preempting
                    # the call to remote, so that it takes data from
                    # the local brick
                    slave.server.xtime = types.MethodType(
                        lambda _self, path, uuid: (
                            brickserver.xtime(path,
                                              uuid + '.' + gconf.slave_id)
                        ),
                        slave.server)
                    slave.server.stime = types.MethodType(
                        lambda _self, path, uuid: (
                            brickserver.stime(path,
                                              uuid + '.' + gconf.slave_id)
                        ),
                        slave.server)
                    slave.server.set_stime = types.MethodType(
                        lambda _self, path, uuid, mark: (
                            brickserver.set_stime(path,
                                                  uuid + '.' + gconf.slave_id,
                                                  mark)
                        ),
                        slave.server)
                (g1, g2, g3) = self.gmaster_instantiate_tuple(slave)
                g1.master.server = brickserver
                g2.master.server = brickserver
                g3.master.server = brickserver
            else:
                (g1, g2, g3) = self.gmaster_instantiate_tuple(slave)
                g1.master.server.aggregated = gmaster.master.server
                g2.master.server.aggregated = gmaster.master.server
                g3.master.server.aggregated = gmaster.master.server
            # bad bad bad: bad way to do things like this
            # need to make this elegant
            # register the crawlers and start crawling
            # g1 ==> Xsync, g2 ==> config.change_detector(changelog by default)
            # g3 ==> changelog History
            changelog_register_failed = False
            (inf, ouf, ra, wa) = gconf.rpc_fd.split(',')
            os.close(int(ra))
            os.close(int(wa))
            changelog_agent = RepceClient(int(inf), int(ouf))
            status = GeorepStatus(gconf.state_file, gconf.local_path)
            status.reset_on_worker_start()
            rv = changelog_agent.version()
            if int(rv) != CHANGELOG_AGENT_CLIENT_VERSION:
                raise GsyncdError(
                    "RePCe major version mismatch(changelog agent): "
                    "local %s, remote %s" %
                    (CHANGELOG_AGENT_CLIENT_VERSION, rv))

            try:
                workdir = g2.setup_working_dir()
                # Register only when change_detector is not set to
                # xsync, else agent will generate changelog files
                # in .processing directory of working dir
                if gconf.change_detector != 'xsync':
                    # register with the changelog library
                    # 9 == log level (DEBUG)
                    # 5 == connection retries
                    changelog_agent.init()
                    changelog_agent.register(gconf.local_path,
                                             workdir, gconf.changelog_log_file,
                                             g2.CHANGELOG_LOG_LEVEL,
                                             g2.CHANGELOG_CONN_RETRIES)

                register_time = int(time.time())
                g2.register(register_time, changelog_agent, status)
                g3.register(register_time, changelog_agent, status)
            except ChangelogException as e:
                logging.error("Changelog register failed, %s" % e)
                sys.exit(1)

            g1.register(status=status)
            logging.info("Register time: %s" % register_time)
            # oneshot: Try to use changelog history api, if not
            # available switch to FS crawl
            # Note: if config.change_detector is xsync then
            # it will not use changelog history api
            try:
                g3.crawlwrap(oneshot=True)
            except PartialHistoryAvailable as e:
                logging.info('Partial history available, using xsync crawl'
                             ' after consuming history till %s' % str(e))
                g1.crawlwrap(oneshot=True, register_time=register_time)
            except NoPurgeTimeAvailable:
                logging.info('No stime available, using xsync crawl')
                g1.crawlwrap(oneshot=True, register_time=register_time)
            except ChangelogException as e:
                logging.error("Changelog History Crawl failed, %s" % e)
                sys.exit(1)

            try:
                g2.crawlwrap()
            except ChangelogException as e:
                logging.error("Changelog crawl failed, %s" % e)
                sys.exit(1)
        else:
            sup(self, *args)

    def rsync(self, files):
        return sup(self, files, self.slavedir)

    def tarssh(self, files):
        return sup(self, files, self.slavedir)


class SSH(AbstractUrl, SlaveRemote):

    """scheme class for ssh:// urls

    interface to remote slave on master side
    implementing an ssh based proxy
    """

    def __init__(self, path):
        self.remote_addr, inner_url = sup(self, path,
                                          '^((?:%s@)?%s):(.+)' %
                                          tuple([r.pattern
                                                 for r in (UserRX, HostRX)]))
        self.inner_rsc = parse_url(inner_url)
        self.volume = inner_url[1:]

    @staticmethod
    def parse_ssh_address(self):
        m = re.match('([^@]+)@(.+)', self.remote_addr)
        if m:
            u, h = m.groups()
        else:
            u, h = syncdutils.getusername(), self.remote_addr
        self.remotehost = h
        return {'user': u, 'host': h}

    def canonical_path(self):
        rap = self.parse_ssh_address(self)
        remote_addr = '@'.join([rap['user'], gethostbyname(rap['host'])])
        return ':'.join([remote_addr, self.inner_rsc.get_url(canonical=True)])

    def can_connect_to(self, remote):
        """determine our position in the connectibility matrix"""
        return False

    def start_fd_client(self, *a, **opts):
        """customizations for client startup

        - be a no-op if we are to daemonize (client startup is deferred
          to post-daemon stage)
        - determine target url for rsync after consulting server
        """
        if opts.get('deferred'):
            return a
        sup(self, *a)
        ityp = type(self.inner_rsc)
        if ityp == FILE:
            slavepath = self.inner_rsc.path
        elif ityp == GLUSTER:
            slavepath = "/proc/%d/cwd" % self.server.pid()
        else:
            raise NotImplementedError
        self.slaveurl = ':'.join([self.remote_addr, slavepath])

    def connect_remote(self, go_daemon=None):
        """connect to inner slave url through outer ssh url

        Wrap the connecting utility in ssh.

        Much care is put into daemonizing: in that case
        ssh is started before daemonization, but
        RePCe client is to be created after that (as ssh
        interactive password auth would be defeated by
        a daemonized ssh, while client should be present
        only in the final process). In that case the action
        is taken apart to two parts, this method is ivoked
        once pre-daemon, once post-daemon. Use @go_daemon
        to deiced what part to perform.

        [NB. ATM gluster product does not makes use of interactive
        authentication.]
        """
        if go_daemon == 'done':
            return self.start_fd_client(*self.fd_pair)

        syncdutils.setup_ssh_ctl(tempfile.mkdtemp(prefix='gsyncd-aux-ssh-'),
                                 self.remote_addr,
                                 self.inner_rsc.url)

        deferred = go_daemon == 'postconn'
        ret = sup(self, gconf.ssh_command.split() + gconf.ssh_ctl_args +
                  [self.remote_addr],
                  slave=self.inner_rsc.url, deferred=deferred)

        if deferred:
            # send a message to peer so that we can wait for
            # the answer from which we know connection is
            # established and we can proceed with daemonization
            # (doing that too early robs the ssh passwd prompt...)
            # However, we'd better not start the RepceClient
            # before daemonization (that's not preserved properly
            # in daemon), we just do a an ad-hoc linear put/get.
            i, o = ret
            inf = os.fdopen(i)
            repce.send(o, None, '__repce_version__')
            select((inf,), (), ())
            repce.recv(inf)
            # hack hack hack: store a global reference to the file
            # to save it from getting GC'd which implies closing it
            gconf.permanent_handles.append(inf)
            self.fd_pair = (i, o)
            return 'should'

    def rsync(self, files):
        return sup(self, files, '-e',
                   " ".join(gconf.ssh_command.split() + gconf.ssh_ctl_args),
                   *(gconf.rsync_ssh_options.split() + [self.slaveurl]))

    def tarssh(self, files):
        return sup(self, files, self.slaveurl)



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

try:
    import ConfigParser
except ImportError:
    # py 3
    import configparser as ConfigParser
import re
from string import Template
import os
import errno
import sys
from stat import ST_DEV, ST_INO, ST_MTIME
import tempfile
import shutil

from syncdutils import escape, unescape, norm, update_file, GsyncdError

SECT_ORD = '__section_order__'
SECT_META = '__meta__'
config_version = 2.0

re_type = type(re.compile(''))


# (SECTION, OPTION, OLD VALUE, NEW VALUE)
CONFIGS = (
    ("peersrx . .",
     "georep_session_working_dir",
     "",
     "/var/lib/glusterd/geo-replication/${mastervol}_${remotehost}_"
     "${slavevol}/"),
    ("peersrx .",
     "gluster_params",
     "aux-gfid-mount xlator-option=\*-dht.assert-no-child-down=true",
     "aux-gfid-mount"),
    ("peersrx .",
     "gluster_params",
     "aux-gfid-mount",
     "aux-gfid-mount acl"),
    ("peersrx . .",
     "ssh_command_tar",
     "",
     "ssh -oPasswordAuthentication=no -oStrictHostKeyChecking=no "
     "-i /var/lib/glusterd/geo-replication/tar_ssh.pem"),
    ("peersrx . .",
     "changelog_log_file",
     "",
     "${iprefix}/log/glusterfs/geo-replication/${mastervol}"
     "/${eSlave}${local_id}-changes.log"),
    ("peersrx . .",
     "working_dir",
     "/var/run/gluster/${mastervol}/${eSlave}",
     "${iprefix}/lib/misc/glusterfsd/${mastervol}/${eSlave}"),
    ("peersrx . .",
     "ignore_deletes",
     "true",
     "false"),
)


def upgrade_config_file(path):
    config_change = False
    config = ConfigParser.RawConfigParser()
    config.read(path)

    for sec, opt, oldval, newval in CONFIGS:
        try:
            val = config.get(sec, opt)
        except ConfigParser.NoOptionError:
            # if new config opt not exists
            config_change = True
            config.set(sec, opt, newval)
            continue
        except ConfigParser.Error:
            """
            When gsyncd invoked at the time of create, config file
            will not be their. Ignore any ConfigParser errors
            """
            continue

        if val == newval:
            # value is same as new val
            continue

        if val == oldval:
            # config value needs update
            config_change = True
            config.set(sec, opt, newval)

    if config_change:
        tempConfigFile = tempfile.NamedTemporaryFile(mode="wb", delete=False)
        with open(tempConfigFile.name, 'wb') as configFile:
            config.write(configFile)

        # If src and dst are two different file system, then os.rename
        # fails, In this case if temp file created in /tmp and if /tmp is
        # separate fs then os.rename gives following error, so use shutil
        # OSError: [Errno 18] Invalid cross-device link
        # mail.python.org/pipermail/python-list/2005-February/342893.html
        shutil.move(tempConfigFile.name, path)


class MultiDict(object):

    """a virtual dict-like class which functions as the union
    of underlying dicts"""

    def __init__(self, *dd):
        self.dicts = dd

    def __getitem__(self, key):
        val = None
        for d in self.dicts:
            if d.get(key) is not None:
                val = d[key]
        if val is None:
            raise KeyError(key)
        return val


class GConffile(object):

    """A high-level interface to ConfigParser which flattens the two-tiered
       config layout by implenting automatic section dispatch based on initial
       parameters.

    Also ensure section ordering in terms of their time of addition -- a compat
    hack for Python < 2.7.
    """

    def _normconfig(self):
        """normalize config keys by s/-/_/g"""
        for n, s in self.config._sections.items():
            if n.find('__') == 0:
                continue
            s2 = type(s)()
            for k, v in s.items():
                if k.find('__') != 0:
                    k = norm(k)
                s2[k] = v
            self.config._sections[n] = s2

    def __init__(self, path, peers, *dd):
        """
        - .path: location of config file
        - .config: underlying ConfigParser instance
        - .peers: on behalf of whom we flatten .config
          (master, or master-slave url pair)
        - .auxdicts: template subtituents
        """
        self.peers = peers
        self.path = path
        self.auxdicts = dd
        self.config = ConfigParser.RawConfigParser()
        self.config.read(path)
        self.dev, self.ino, self.mtime = -1, -1, -1
        self._normconfig()

    def _load(self):
        try:
            sres = os.stat(self.path)
            self.dev = sres[ST_DEV]
            self.ino = sres[ST_INO]
            self.mtime = sres[ST_MTIME]
        except (OSError, IOError):
            if sys.exc_info()[1].errno == errno.ENOENT:
                sres = None

        self.config = ConfigParser.RawConfigParser()
        self.config.read(self.path)
        self._normconfig()

    def get_realtime(self, opt):
        try:
            sres = os.stat(self.path)
        except (OSError, IOError):
            if sys.exc_info()[1].errno == errno.ENOENT:
                sres = None
            else:
                raise

        # compare file system stat with that of our stream file handle
        if not sres or sres[ST_DEV] != self.dev or \
           sres[ST_INO] != self.ino or self.mtime != sres[ST_MTIME]:
            self._load()

        return self.get(opt, printValue=False)

    def section(self, rx=False):
        """get the section name of the section representing .peers
        in .config"""
        peers = self.peers
        if not peers:
            peers = ['.', '.']
            rx = True
        if rx:
            st = 'peersrx'
        else:
            st = 'peers'
        return ' '.join([st] + [escape(u) for u in peers])

    @staticmethod
    def parse_section(section):
        """retrieve peers sequence encoded by section name
           (as urls or regexen, depending on section type)
        """
        sl = section.split()
        st = sl.pop(0)
        sl = [unescape(u) for u in sl]
        if st == 'peersrx':
            sl = [re.compile(u) for u in sl]
        return sl

    def ord_sections(self):
        """Return an ordered list of sections.

        Ordering happens based on the auxiliary
        SECT_ORD section storing indices for each
        section added through the config API.

        To not to go corrupt in case of manually
        written config files, we take care to append
        also those sections which are not registered
        in SECT_ORD.

        Needed for python 2.{4,5,6} where ConfigParser
        cannot yet order sections/options internally.
        """
        so = {}
        if self.config.has_section(SECT_ORD):
            so = self.config._sections[SECT_ORD]
        so2 = {}
        for k, v in so.items():
            if k != '__name__':
                so2[k] = int(v)
        tv = 0
        if so2:
            tv = max(so2.values()) + 1
        ss = [s for s in self.config.sections() if s.find('__') != 0]
        for s in ss:
            if s in so.keys():
                continue
            so2[s] = tv
            tv += 1

        def scmp(x, y):
            return cmp(*(so2[s] for s in (x, y)))
        ss.sort(scmp)
        return ss

    def update_to(self, dct, allow_unresolved=False):
        """update @dct from key/values of ours.

        key/values are collected from .config by filtering the regexp sections
        according to match, and from .section. The values are treated as
        templates, which are substituted from .auxdicts and (in case of regexp
        sections) match groups.
        """
        if not self.peers:
            raise GsyncdError('no peers given, cannot select matching options')

        def update_from_sect(sect, mud):
            for k, v in self.config._sections[sect].items():
                if k == '__name__':
                    continue
                if allow_unresolved:
                    dct[k] = Template(v).safe_substitute(mud)
                else:
                    dct[k] = Template(v).substitute(mud)
        for sect in self.ord_sections():
            sp = self.parse_section(sect)
            if isinstance(sp[0], re_type) and len(sp) == len(self.peers):
                match = True
                mad = {}
                for i in range(len(sp)):
                    m = sp[i].search(self.peers[i])
                    if not m:
                        match = False
                        break
                    for j in range(len(m.groups())):
                        mad['match%d_%d' % (i + 1, j + 1)] = m.groups()[j]
                if match:
                    update_from_sect(sect, MultiDict(dct, mad, *self.auxdicts))
        if self.config.has_section(self.section()):
            update_from_sect(self.section(), MultiDict(dct, *self.auxdicts))

    def get(self, opt=None, printValue=True):
        """print the matching key/value pairs from .config,
           or if @opt given, the value for @opt (according to the
           logic described in .update_to)
        """
        d = {}
        self.update_to(d, allow_unresolved=True)
        if opt:
            opt = norm(opt)
            v = d.get(opt)
            if v:
                if printValue:
                    print(v)
                else:
                    return v
        else:
            for k, v in d.iteritems():
                if k == '__name__':
                    continue
                print("%s: %s" % (k, v))

    def write(self, trfn, opt, *a, **kw):
        """update on-disk config transactionally

        @trfn is the transaction function
        """
        def mergeconf(f):
            self.config = ConfigParser.RawConfigParser()
            self.config.readfp(f)
            self._normconfig()
            if not self.config.has_section(SECT_META):
                self.config.add_section(SECT_META)
            self.config.set(SECT_META, 'version', config_version)
            return trfn(norm(opt), *a, **kw)

        def updateconf(f):
            self.config.write(f)
        update_file(self.path, updateconf, mergeconf)

    def _set(self, opt, val, rx=False):
        """set @opt to @val in .section"""
        sect = self.section(rx)
        if not self.config.has_section(sect):
            self.config.add_section(sect)
            # regarding SECT_ORD, cf. ord_sections
            if not self.config.has_section(SECT_ORD):
                self.config.add_section(SECT_ORD)
            self.config.set(
                SECT_ORD, sect, len(self.config._sections[SECT_ORD]))
        self.config.set(sect, opt, val)
        return True

    def set(self, opt, *a, **kw):
        """perform ._set transactionally"""
        self.write(self._set, opt, *a, **kw)

    def _delete(self, opt, rx=False):
        """delete @opt from .section"""
        sect = self.section(rx)
        if self.config.has_section(sect):
            return self.config.remove_option(sect, opt)

    def delete(self, opt, *a, **kw):
        """perform ._delete transactionally"""
        self.write(self._delete, opt, *a, **kw)



#!/usr/bin/env python
#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import fcntl
import os
import tempfile
import urllib
import json
import time
from datetime import datetime

DEFAULT_STATUS = "N/A"
MONITOR_STATUS = ("Created", "Started", "Paused", "Stopped")
STATUS_VALUES = (DEFAULT_STATUS,
                 "Initializing...",
                 "Active",
                 "Passive",
                 "Faulty")

CRAWL_STATUS_VALUES = (DEFAULT_STATUS,
                       "Hybrid Crawl",
                       "History Crawl",
                       "Changelog Crawl")


def human_time(ts):
    try:
        return datetime.fromtimestamp(float(ts)).strftime("%Y-%m-%d %H:%M:%S")
    except ValueError:
        return DEFAULT_STATUS


def human_time_utc(ts):
    try:
        return datetime.utcfromtimestamp(
            float(ts)).strftime("%Y-%m-%d %H:%M:%S")
    except ValueError:
        return DEFAULT_STATUS


def get_default_values():
    return {
        "slave_node": DEFAULT_STATUS,
        "worker_status": DEFAULT_STATUS,
        "last_synced": 0,
        "crawl_status": DEFAULT_STATUS,
        "entry": 0,
        "data": 0,
        "meta": 0,
        "failures": 0,
        "checkpoint_completed": DEFAULT_STATUS,
        "checkpoint_time": 0,
        "checkpoint_completion_time": 0}


class LockedOpen(object):

    def __init__(self, filename, *args, **kwargs):
        self.filename = filename
        self.open_args = args
        self.open_kwargs = kwargs
        self.fileobj = None

    def __enter__(self):
        """
        If two processes compete to update a file, The first process
        gets the lock and the second process is blocked in the fcntl.flock()
        call. When first process replaces the file and releases the lock,
        the already open file descriptor in the second process now points
        to a  "ghost" file(not reachable by any path name) with old contents.
        To avoid that conflict, check the fd already opened is same or
        not. Open new one if not same
        """
        f = open(self.filename, *self.open_args, **self.open_kwargs)
        while True:
            fcntl.flock(f, fcntl.LOCK_EX)
            fnew = open(self.filename, *self.open_args, **self.open_kwargs)
            if os.path.sameopenfile(f.fileno(), fnew.fileno()):
                fnew.close()
                break
            else:
                f.close()
                f = fnew
        self.fileobj = f
        return f

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.fileobj.close()


def set_monitor_status(status_file, status):
    fd = os.open(status_file, os.O_CREAT | os.O_RDWR)
    os.close(fd)
    with LockedOpen(status_file, 'r+'):
        with tempfile.NamedTemporaryFile('w', dir=os.path.dirname(status_file),
                                         delete=False) as tf:
            tf.write(status)
            tempname = tf.name

        os.rename(tempname, status_file)
        dirfd = os.open(os.path.dirname(os.path.abspath(status_file)),
                        os.O_DIRECTORY)
        os.fsync(dirfd)
        os.close(dirfd)


class GeorepStatus(object):
    def __init__(self, monitor_status_file, brick):
        self.work_dir = os.path.dirname(monitor_status_file)
        self.monitor_status_file = monitor_status_file
        self.filename = os.path.join(self.work_dir,
                                     "brick_%s.status"
                                     % urllib.quote_plus(brick))

        fd = os.open(self.filename, os.O_CREAT | os.O_RDWR)
        os.close(fd)
        fd = os.open(self.monitor_status_file, os.O_CREAT | os.O_RDWR)
        os.close(fd)
        self.brick = brick
        self.default_values = get_default_values()

    def _update(self, mergerfunc):
        with LockedOpen(self.filename, 'r+') as f:
            try:
                data = json.load(f)
            except ValueError:
                data = self.default_values

            data = mergerfunc(data)
            with tempfile.NamedTemporaryFile(
                    'w',
                    dir=os.path.dirname(self.filename),
                    delete=False) as tf:
                tf.write(data)
                tempname = tf.name

            os.rename(tempname, self.filename)
            dirfd = os.open(os.path.dirname(os.path.abspath(self.filename)),
                            os.O_DIRECTORY)
            os.fsync(dirfd)
            os.close(dirfd)

    def reset_on_worker_start(self):
        def merger(data):
            data["slave_node"] = DEFAULT_STATUS
            data["crawl_status"] = DEFAULT_STATUS
            data["entry"] = 0
            data["data"] = 0
            data["meta"] = 0
            return json.dumps(data)

        self._update(merger)

    def set_field(self, key, value):
        def merger(data):
            data[key] = value
            return json.dumps(data)

        self._update(merger)

    def set_last_synced(self, value, checkpoint_time):
        def merger(data):
            data["last_synced"] = value[0]

            # If checkpoint is not set or reset
            # or if last set checkpoint is changed
            if checkpoint_time == 0 or \
               checkpoint_time != data["checkpoint_time"]:
                data["checkpoint_time"] = 0
                data["checkpoint_completion_time"] = 0
                data["checkpoint_completed"] = "No"

            # If checkpoint is completed and not marked as completed
            # previously then update the checkpoint completed time
            if checkpoint_time > 0 and checkpoint_time <= value[0]:
                if data["checkpoint_completed"] == "No":
                    data["checkpoint_time"] = checkpoint_time
                    data["checkpoint_completion_time"] = int(time.time())
                    data["checkpoint_completed"] = "Yes"
            return json.dumps(data)

        self._update(merger)

    def set_worker_status(self, status):
        self.set_field("worker_status", status)

    def set_worker_crawl_status(self, status):
        self.set_field("crawl_status", status)

    def set_slave_node(self, slave_node):
        def merger(data):
            data["slave_node"] = slave_node
            return json.dumps(data)

        self._update(merger)

    def inc_value(self, key, value):
        def merger(data):
            data[key] = data.get(key, 0) + value
            return json.dumps(data)

        self._update(merger)

    def dec_value(self, key, value):
        def merger(data):
            data[key] = data.get(key, 0) - value
            if data[key] < 0:
                data[key] = 0
            return json.dumps(data)

        self._update(merger)

    def set_active(self):
        self.set_field("worker_status", "Active")

    def set_passive(self):
        self.set_field("worker_status", "Passive")

    def get_monitor_status(self):
        data = ""
        with open(self.monitor_status_file, "r") as f:
            data = f.read().strip()
        return data

    def get_status(self, checkpoint_time=0):
        """
        Monitor Status --->        Created    Started  Paused      Stopped
        ----------------------------------------------------------------------
        slave_node                 N/A        VALUE    VALUE       N/A
        status                     Created    VALUE    Paused      Stopped
        last_synced                N/A        VALUE    VALUE       VALUE
        crawl_status               N/A        VALUE    N/A         N/A
        entry                      N/A        VALUE    N/A         N/A
        data                       N/A        VALUE    N/A         N/A
        meta                       N/A        VALUE    N/A         N/A
        failures                   N/A        VALUE    VALUE       VALUE
        checkpoint_completed       N/A        VALUE    VALUE       VALUE
        checkpoint_time            N/A        VALUE    VALUE       VALUE
        checkpoint_completed_time  N/A        VALUE    VALUE       VALUE
        """
        data = self.default_values
        with open(self.filename) as f:
            try:
                data.update(json.load(f))
            except ValueError:
                pass
        monitor_status = self.get_monitor_status()

        if monitor_status in ["Created", "Paused", "Stopped"]:
            data["worker_status"] = monitor_status

        # Checkpoint adjustments
        if checkpoint_time == 0:
            data["checkpoint_completed"] = DEFAULT_STATUS
            data["checkpoint_time"] = DEFAULT_STATUS
            data["checkpoint_completion_time"] = DEFAULT_STATUS
        else:
            if checkpoint_time != data["checkpoint_time"]:
                if checkpoint_time <= data["last_synced"]:
                    data["checkpoint_completed"] = "Yes"
                    data["checkpoint_time"] = checkpoint_time
                    data["checkpoint_completion_time"] = data["last_synced"]
                else:
                    data["checkpoint_completed"] = "No"
                    data["checkpoint_time"] = checkpoint_time
                    data["checkpoint_completion_time"] = DEFAULT_STATUS

        if data["checkpoint_time"] not in [0, DEFAULT_STATUS]:
            chkpt_time = data["checkpoint_time"]
            data["checkpoint_time"] = human_time(chkpt_time)
            data["checkpoint_time_utc"] = human_time_utc(chkpt_time)

        if data["checkpoint_completion_time"] not in [0, DEFAULT_STATUS]:
            chkpt_completion_time = data["checkpoint_completion_time"]
            data["checkpoint_completion_time"] = human_time(
                chkpt_completion_time)
            data["checkpoint_completion_time_utc"] = human_time_utc(
                chkpt_completion_time)

        if data["last_synced"] == 0:
            data["last_synced"] = DEFAULT_STATUS
            data["last_synced_utc"] = DEFAULT_STATUS
        else:
            last_synced = data["last_synced"]
            data["last_synced"] = human_time(last_synced)
            data["last_synced_utc"] = human_time_utc(last_synced)

        if data["worker_status"] != "Active":
            data["last_synced"] = DEFAULT_STATUS
            data["last_synced_utc"] = DEFAULT_STATUS
            data["crawl_status"] = DEFAULT_STATUS
            data["entry"] = DEFAULT_STATUS
            data["data"] = DEFAULT_STATUS
            data["meta"] = DEFAULT_STATUS
            data["failures"] = DEFAULT_STATUS
            data["checkpoint_completed"] = DEFAULT_STATUS
            data["checkpoint_time"] = DEFAULT_STATUS
            data["checkpoint_completed_time"] = DEFAULT_STATUS
            data["checkpoint_time_utc"] = DEFAULT_STATUS
            data["checkpoint_completion_time_utc"] = DEFAULT_STATUS

        if data["worker_status"] not in ["Active", "Passive"]:
            data["slave_node"] = DEFAULT_STATUS

        if data.get("last_synced_utc", 0) == 0:
            data["last_synced_utc"] = DEFAULT_STATUS

        if data.get("checkpoint_completion_time_utc", 0) == 0:
            data["checkpoint_completion_time_utc"] = DEFAULT_STATUS

        if data.get("checkpoint_time_utc", 0) == 0:
            data["checkpoint_time_utc"] = DEFAULT_STATUS

        return data

    def print_status(self, checkpoint_time=0):
        for key, value in self.get_status(checkpoint_time).items():
            print ("%s: %s" % (key, value))



import sys
import stat
from uuid import UUID
from time import strftime, localtime
from gluster.glupy import *

# This translator was written primarily to test the fop entry point definitions
# and structure definitions in 'glupy.py'.

# It is similar to the C language debug-trace translator, which logs the
# arguments passed to the fops and their corresponding cbk functions.

dl.get_id.restype = c_long
dl.get_id.argtypes = [ POINTER(call_frame_t) ]

dl.get_rootunique.restype = c_uint64
dl.get_rootunique.argtypes = [ POINTER(call_frame_t) ]

def uuid2str (gfid):
        return str(UUID(''.join(map("{0:02x}".format, gfid))))


def st_mode_from_ia (prot, filetype):
        st_mode = 0
        type_bit = 0
        prot_bit = 0

        if filetype == IA_IFREG:
                type_bit = stat.S_IFREG
        elif filetype == IA_IFDIR:
                type_bit = stat.S_IFDIR
        elif filetype == IA_IFLNK:
                type_bit = stat.S_IFLNK
        elif filetype == IA_IFBLK:
                type_bit = stat.S_IFBLK
        elif filetype == IA_IFCHR:
                type_bit = stat.S_IFCHR
        elif filetype == IA_IFIFO:
                type_bit = stat.S_IFIFO
        elif filetype == IA_IFSOCK:
                type_bit = stat.S_IFSOCK
        elif filetype == IA_INVAL:
                pass


        if prot.suid:
                prot_bit |= stat.S_ISUID
        if prot.sgid:
                prot_bit |= stat.S_ISGID
        if prot.sticky:
                prot_bit |= stat.S_ISVTX

        if prot.owner.read:
                prot_bit |= stat.S_IRUSR
        if prot.owner.write:
                prot_bit |= stat.S_IWUSR
        if prot.owner.execn:
                prot_bit |= stat.S_IXUSR

        if prot.group.read:
                prot_bit |= stat.S_IRGRP
        if prot.group.write:
                prot_bit |= stat.S_IWGRP
        if prot.group.execn:
                prot_bit |= stat.S_IXGRP

        if prot.other.read:
                prot_bit |= stat.S_IROTH
        if prot.other.write:
                prot_bit |= stat.S_IWOTH
        if prot.other.execn:
                prot_bit |= stat.S_IXOTH

        st_mode = (type_bit | prot_bit)

        return st_mode


def trace_stat2str (buf):
        gfid = uuid2str(buf.contents.ia_gfid)
        mode = st_mode_from_ia(buf.contents.ia_prot, buf.contents.ia_type)
        atime_buf = strftime("[%b %d %H:%M:%S]",
                             localtime(buf.contents.ia_atime))
        mtime_buf = strftime("[%b %d %H:%M:%S]",
                             localtime(buf.contents.ia_mtime))
        ctime_buf = strftime("[%b %d %H:%M:%S]",
                             localtime(buf.contents.ia_ctime))
        return ("(gfid={0:s}, ino={1:d}, mode={2:o}, nlink={3:d}, uid ={4:d}, "+
                "gid ={5:d}, size={6:d}, blocks={7:d}, atime={8:s}, mtime={9:s}, "+
                "ctime={10:s})").format(gfid, buf.contents.ia_no, mode,
                                        buf.contents.ia_nlink,
                                        buf.contents.ia_uid,
                                        buf.contents.ia_gid,
                                        buf.contents.ia_size,
                                        buf.contents.ia_blocks,
                                        atime_buf, mtime_buf,
                                        ctime_buf)

class xlator(Translator):

        def __init__(self, c_this):
                Translator.__init__(self, c_this)
                self.gfids = {}

        def lookup_fop(self, frame, this, loc, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.gfid)
                print("GLUPY TRACE LOOKUP FOP- {0:d}: gfid={1:s}; " +
                      "path={2:s}").format(unique, gfid, loc.contents.path)
                self.gfids[key] = gfid
                dl.wind_lookup(frame, POINTER(xlator_t)(), loc, xdata)
                return 0

        def lookup_cbk(self, frame, cookie, this, op_ret, op_errno,
                       inode, buf, xdata, postparent):
                unique =dl.get_rootunique(frame)
                key =dl.get_id(frame)
                if op_ret == 0:
                        gfid = uuid2str(buf.contents.ia_gfid)
                        statstr = trace_stat2str(buf)
                        postparentstr = trace_stat2str(postparent)
                        print("GLUPY TRACE LOOKUP CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d}; *buf={3:s}; " +
                              "*postparent={4:s}").format(unique, gfid,
                                                          op_ret, statstr,
                                                          postparentstr)
                else:
                        gfid = self.gfids[key]
                        print("GLUPY TRACE LOOKUP CBK - {0:d}: gfid={1:s};" +
                              " op_ret={2:d}; op_errno={3:d}").format(unique,
                                                                      gfid,
                                                                      op_ret,
                                                                      op_errno)
                del self.gfids[key]
                dl.unwind_lookup(frame, cookie, this, op_ret, op_errno,
                                 inode, buf, xdata, postparent)
                return 0

        def create_fop(self, frame, this, loc, flags, mode, umask, fd,
                       xdata):
                unique = dl.get_rootunique(frame)
                gfid = uuid2str(loc.contents.gfid)
                print("GLUPY TRACE CREATE FOP- {0:d}: gfid={1:s}; path={2:s}; " +
                      "fd={3:s}; flags=0{4:o}; mode=0{5:o}; " +
                      "umask=0{6:o}").format(unique, gfid, loc.contents.path,
                                             fd, flags, mode, umask)
                dl.wind_create(frame, POINTER(xlator_t)(), loc, flags,mode,
                               umask, fd, xdata)
                return 0

        def create_cbk(self, frame, cookie, this, op_ret, op_errno, fd,
                       inode, buf, preparent, postparent, xdata):
                unique = dl.get_rootunique(frame)
                if op_ret >= 0:
                        gfid = uuid2str(inode.contents.gfid)
                        statstr = trace_stat2str(buf)
                        preparentstr = trace_stat2str(preparent)
                        postparentstr = trace_stat2str(postparent)
                        print("GLUPY TRACE CREATE CBK- {0:d}: gfid={1:s};" +
                              " op_ret={2:d}; fd={3:s}; *stbuf={4:s}; " +
                              "*preparent={5:s};" +
                              " *postparent={6:s}").format(unique, gfid, op_ret,
                                                           fd, statstr,
                                                           preparentstr,
                                                           postparentstr)
                else:
                        print ("GLUPY TRACE CREATE CBK- {0:d}: op_ret={1:d}; " +
                              "op_errno={2:d}").format(unique, op_ret, op_errno)
                dl.unwind_create(frame, cookie, this, op_ret, op_errno, fd,
                                 inode, buf, preparent, postparent, xdata)
                return 0

        def open_fop(self, frame, this, loc, flags, fd, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE OPEN FOP- {0:d}: gfid={1:s}; path={2:s}; "+
                      "flags={3:d}; fd={4:s}").format(unique, gfid,
                                                      loc.contents.path, flags,
                                                      fd)
                self.gfids[key] = gfid
                dl.wind_open(frame, POINTER(xlator_t)(), loc, flags, fd, xdata)
                return 0

        def open_cbk(self, frame, cookie, this, op_ret, op_errno, fd, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                print("GLUPY TRACE OPEN CBK- {0:d}: gfid={1:s}; op_ret={2:d}; "
                      "op_errno={3:d}; *fd={4:s}").format(unique, gfid,
                                                          op_ret, op_errno, fd)
                del self.gfids[key]
                dl.unwind_open(frame, cookie, this, op_ret, op_errno, fd,
                               xdata)
                return 0

        def readv_fop(self, frame, this, fd, size, offset, flags, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(fd.contents.inode.contents.gfid)
                print("GLUPY TRACE READV FOP- {0:d}: gfid={1:s}; "+
                      "fd={2:s}; size ={3:d}; offset={4:d}; " +
                      "flags=0{5:x}").format(unique, gfid, fd, size, offset,
                                             flags)
                self.gfids[key] = gfid
                dl.wind_readv (frame, POINTER(xlator_t)(), fd, size, offset,
                               flags, xdata)
                return 0

        def readv_cbk(self, frame, cookie, this, op_ret, op_errno, vector,
                      count, buf, iobref, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                if op_ret >= 0:
                        statstr = trace_stat2str(buf)
                        print("GLUPY TRACE READV CBK- {0:d}: gfid={1:s}, "+
                              "op_ret={2:d}; *buf={3:s};").format(unique, gfid,
                                                                  op_ret,
                                                                  statstr)

                else:
                        print("GLUPY TRACE READV CBK- {0:d}: gfid={1:s}, "+
                              "op_ret={2:d}; op_errno={3:d}").format(unique,
                                                                     gfid,
                                                                     op_ret,
                                                                     op_errno)
                del self.gfids[key]
                dl.unwind_readv (frame, cookie, this, op_ret, op_errno,
                                 vector, count, buf, iobref, xdata)
                return 0

        def writev_fop(self, frame, this, fd, vector, count, offset, flags,
                       iobref, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(fd.contents.inode.contents.gfid)
                print("GLUPY TRACE  WRITEV FOP- {0:d}: gfid={1:s}; " +
                      "fd={2:s}; count={3:d}; offset={4:d}; " +
                      "flags=0{5:x}").format(unique, gfid, fd, count, offset,
                                             flags)
                self.gfids[key] = gfid
                dl.wind_writev(frame, POINTER(xlator_t)(), fd, vector, count,
                               offset, flags, iobref, xdata)
                return 0

        def writev_cbk(self, frame, cookie, this, op_ret, op_errno, prebuf,
                       postbuf, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                if op_ret >= 0:
                        preopstr = trace_stat2str(prebuf)
                        postopstr = trace_stat2str(postbuf)
                        print("GLUPY TRACE WRITEV CBK- {0:d}: op_ret={1:d}; " +
                              "*prebuf={2:s}; " +
                              "*postbuf={3:s}").format(unique, op_ret, preopstr,
                                                       postopstr)
                else:
                        gfid = self.gfids[key]
                        print("GLUPY TRACE WRITEV CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d}; op_errno={3:d}").format(unique,
                                                                     gfid,
                                                                     op_ret,
                                                                     op_errno)
                del self.gfids[key]
                dl.unwind_writev (frame, cookie, this, op_ret, op_errno,
                                  prebuf, postbuf, xdata)
                return 0

        def opendir_fop(self, frame, this, loc, fd, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE OPENDIR FOP- {0:d}: gfid={1:s}; path={2:s}; "+
                      "fd={3:s}").format(unique, gfid, loc.contents.path, fd)
                self.gfids[key] = gfid
                dl.wind_opendir(frame, POINTER(xlator_t)(), loc, fd, xdata)
                return 0

        def opendir_cbk(self, frame, cookie, this, op_ret, op_errno, fd,
                        xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                print("GLUPY TRACE OPENDIR CBK- {0:d}: gfid={1:s}; op_ret={2:d};"+
                      " op_errno={3:d}; fd={4:s}").format(unique, gfid, op_ret,
                                                          op_errno, fd)
                del self.gfids[key]
                dl.unwind_opendir(frame, cookie, this, op_ret, op_errno,
                                  fd, xdata)
                return 0

        def readdir_fop(self, frame, this, fd, size, offset, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(fd.contents.inode.contents.gfid)
                print("GLUPY TRACE READDIR FOP- {0:d}:  gfid={1:s}; fd={2:s}; " +
                      "size={3:d}; offset={4:d}").format(unique, gfid, fd, size,
                                                         offset)
                self.gfids[key] = gfid
                dl.wind_readdir(frame, POINTER(xlator_t)(), fd, size, offset,
                                xdata)
                return 0

        def readdir_cbk(self, frame, cookie, this, op_ret, op_errno, buf,
                        xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                print("GLUPY TRACE READDIR CBK- {0:d}: gfid={1:s}; op_ret={2:d};"+
                      " op_errno={3:d}").format(unique, gfid, op_ret, op_errno)
                del self.gfids[key]
                dl.unwind_readdir(frame, cookie, this, op_ret, op_errno, buf,
                                  xdata)
                return 0

        def readdirp_fop(self, frame, this, fd, size, offset, dictionary):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(fd.contents.inode.contents.gfid)
                print("GLUPY TRACE READDIRP FOP- {0:d}: gfid={1:s}; fd={2:s}; "+
                      " size={3:d}; offset={4:d}").format(unique, gfid, fd, size,
                                                          offset)
                self.gfids[key] = gfid
                dl.wind_readdirp(frame, POINTER(xlator_t)(), fd, size, offset,
                                 dictionary)
                return 0

        def readdirp_cbk(self, frame, cookie, this, op_ret, op_errno, buf,
                         xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                print("GLUPY TRACE READDIRP CBK- {0:d}: gfid={1:s}; "+
                      "op_ret={2:d}; op_errno={3:d}").format(unique, gfid,
                                                             op_ret, op_errno)
                del self.gfids[key]
                dl.unwind_readdirp(frame, cookie, this, op_ret, op_errno, buf,
                                  xdata)
                return 0

        def mkdir_fop(self, frame, this, loc, mode, umask, xdata):
                unique = dl.get_rootunique(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE MKDIR FOP- {0:d}: gfid={1:s}; path={2:s}; " +
                      "mode={3:d}; umask=0{4:o}").format(unique, gfid,
                                                         loc.contents.path, mode,
                                                         umask)
                dl.wind_mkdir(frame, POINTER(xlator_t)(), loc, mode, umask,
                              xdata)
                return 0

        def mkdir_cbk(self, frame, cookie, this, op_ret, op_errno, inode, buf,
                      preparent, postparent,  xdata):
                unique = dl.get_rootunique(frame)
                if op_ret == 0:
                        gfid = uuid2str(inode.contents.gfid)
                        statstr = trace_stat2str(buf)
                        preparentstr = trace_stat2str(preparent)
                        postparentstr = trace_stat2str(postparent)
                        print("GLUPY TRACE MKDIR CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d}; *stbuf={3:s}; *prebuf={4:s}; "+
                              "*postbuf={5:s} ").format(unique, gfid, op_ret,
                                                        statstr,
                                                        preparentstr,
                                                        postparentstr)
                else:
                        print("GLUPY TRACE MKDIR CBK- {0:d}:  op_ret={1:d}; "+
                              "op_errno={2:d}").format(unique, op_ret, op_errno)
                dl.unwind_mkdir(frame, cookie, this, op_ret, op_errno, inode,
                                buf, preparent, postparent, xdata)
                return 0

        def rmdir_fop(self, frame, this, loc, flags, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE RMDIR FOP- {0:d}: gfid={1:s}; path={2:s}; "+
                      "flags={3:d}").format(unique, gfid, loc.contents.path,
                                            flags)
                self.gfids[key] = gfid
                dl.wind_rmdir(frame, POINTER(xlator_t)(), loc, flags, xdata)
                return 0

        def rmdir_cbk(self, frame, cookie, this, op_ret, op_errno, preparent,
                      postparent, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                if op_ret == 0:
                        preparentstr = trace_stat2str(preparent)
                        postparentstr = trace_stat2str(postparent)
                        print("GLUPY TRACE RMDIR CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d}; *prebuf={3:s}; "+
                              "*postbuf={4:s}").format(unique, gfid, op_ret,
                                                       preparentstr,
                                                       postparentstr)
                else:
                        print("GLUPY TRACE RMDIR CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d}; op_errno={3:d}").format(unique,
                                                                     gfid,
                                                                     op_ret,
                                                                     op_errno)
                del self.gfids[key]
                dl.unwind_rmdir(frame, cookie, this, op_ret, op_errno,
                                preparent, postparent, xdata)
                return 0

        def stat_fop(self, frame, this, loc, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE STAT FOP- {0:d}: gfid={1:s}; " +
                      " path={2:s}").format(unique, gfid, loc.contents.path)
                self.gfids[key] = gfid
                dl.wind_stat(frame, POINTER(xlator_t)(), loc, xdata)
                return 0

        def stat_cbk(self, frame, cookie, this, op_ret, op_errno, buf,
                     xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                if op_ret == 0:
                        statstr = trace_stat2str(buf)
                        print("GLUPY TRACE STAT CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d};  *buf={3:s};").format(unique,
                                                                   gfid,
                                                                   op_ret,
                                                                   statstr)
                else:
                        print("GLUPY TRACE STAT CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d}; op_errno={3:d}").format(unique,
                                                                     gfid,
                                                                     op_ret,
                                                                     op_errno)
                del self.gfids[key]
                dl.unwind_stat(frame, cookie, this, op_ret, op_errno,
                               buf, xdata)
                return 0

        def fstat_fop(self, frame, this, fd, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(fd.contents.inode.contents.gfid)
                print("GLUPY TRACE FSTAT FOP- {0:d}:  gfid={1:s}; " +
                      "fd={2:s}").format(unique, gfid, fd)
                self.gfids[key] = gfid
                dl.wind_fstat(frame, POINTER(xlator_t)(), fd, xdata)
                return 0

        def fstat_cbk(self, frame, cookie, this, op_ret, op_errno, buf,
                      xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                if op_ret == 0:
                        statstr = trace_stat2str(buf)
                        print("GLUPY TRACE FSTAT CBK- {0:d}: gfid={1:s} "+
                              " op_ret={2:d}; *buf={3:s}").format(unique,
                                                                  gfid,
                                                                  op_ret,
                                                                  statstr)
                else:
                        print("GLUPY TRACE FSTAT CBK- {0:d}: gfid={1:s} "+
                              "op_ret={2:d}; op_errno={3:d}").format(unique.
                                                                     gfid,
                                                                     op_ret,
                                                                     op_errno)
                del self.gfids[key]
                dl.unwind_fstat(frame, cookie, this, op_ret, op_errno,
                                buf, xdata)
                return 0

        def statfs_fop(self, frame, this, loc, xdata):
                unique = dl.get_rootunique(frame)
                if loc.contents.inode:
                        gfid = uuid2str(loc.contents.inode.contents.gfid)
                else:
                        gfid = "0"
                print("GLUPY TRACE STATFS FOP- {0:d}: gfid={1:s}; "+
                      "path={2:s}").format(unique, gfid, loc.contents.path)
                dl.wind_statfs(frame, POINTER(xlator_t)(), loc, xdata)
                return 0

        def statfs_cbk(self, frame, cookie, this, op_ret, op_errno, buf,
                       xdata):
                unique = dl.get_rootunique(frame)
                if op_ret == 0:
                        #TBD: print buf (pointer to an iovec type object)
                        print("GLUPY TRACE STATFS CBK {0:d}: "+
                              "op_ret={1:d}").format(unique, op_ret)
                else:
                        print("GLUPY TRACE STATFS CBK-  {0:d}"+
                              "op_ret={1:d}; op_errno={2:d}").format(unique,
                                                                     op_ret,
                                                                     op_errno)
                dl.unwind_statfs(frame, cookie, this, op_ret, op_errno,
                                 buf, xdata)
                return 0

        def getxattr_fop(self, frame, this, loc, name, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE GETXATTR FOP- {0:d}: gfid={1:s}; path={2:s};"+
                      " name={3:s}").format(unique, gfid, loc.contents.path,
                                            name)
                self.gfids[key]=gfid
                dl.wind_getxattr(frame, POINTER(xlator_t)(), loc, name, xdata)
                return 0

        def getxattr_cbk(self, frame, cookie, this, op_ret, op_errno,
                         dictionary, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                print("GLUPY TRACE GETXATTR CBK- {0:d}: gfid={1:s}; "+
                      "op_ret={2:d}; op_errno={3:d}; "+
                      " dictionary={4:s}").format(unique, gfid, op_ret, op_errno,
                                                  dictionary)
                del self.gfids[key]
                dl.unwind_getxattr(frame, cookie, this, op_ret, op_errno,
                                   dictionary, xdata)
                return 0

        def fgetxattr_fop(self, frame, this, fd, name, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(fd.contents.inode.contents.gfid)
                print("GLUPY TRACE FGETXATTR FOP- {0:d}: gfid={1:s}; fd={2:s}; "+
                      "name={3:s}").format(unique, gfid, fd, name)
                self.gfids[key] = gfid
                dl.wind_fgetxattr(frame, POINTER(xlator_t)(), fd, name, xdata)
                return 0

        def fgetxattr_cbk(self, frame, cookie, this, op_ret, op_errno,
                          dictionary, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                print("GLUPY TRACE FGETXATTR CBK- {0:d}: gfid={1:s}; "+
                      "op_ret={2:d}; op_errno={3:d};"+
                      " dictionary={4:s}").format(unique, gfid, op_ret,
                                                  op_errno, dictionary)
                del self.gfids[key]
                dl.unwind_fgetxattr(frame, cookie, this, op_ret, op_errno,
                                    dictionary, xdata)
                return 0

        def setxattr_fop(self, frame, this, loc, dictionary, flags, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE SETXATTR FOP- {0:d}:  gfid={1:s}; path={2:s};"+
                      " flags={3:d}").format(unique, gfid, loc.contents.path,
                                             flags)
                self.gfids[key] = gfid
                dl.wind_setxattr(frame, POINTER(xlator_t)(), loc, dictionary,
                                 flags, xdata)
                return 0

        def setxattr_cbk(self, frame, cookie, this, op_ret, op_errno, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                print("GLUPY TRACE SETXATTR CBK- {0:d}: gfid={1:s}; "+
                      "op_ret={2:d}; op_errno={3:d}").format(unique, gfid,
                                                             op_ret, op_errno)
                del self.gfids[key]
                dl.unwind_setxattr(frame, cookie, this, op_ret, op_errno,
                                   xdata)
                return 0

        def fsetxattr_fop(self, frame, this, fd, dictionary, flags, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(fd.contents.inode.contents.gfid)
                print("GLUPY TRACE FSETXATTR FOP- {0:d}: gfid={1:s}; fd={2:p}; "+
                      "flags={3:d}").format(unique, gfid, fd, flags)
                self.gfids[key] = gfid
                dl.wind_fsetxattr(frame, POINTER(xlator_t)(), fd, dictionary,
                                  flags, xdata)
                return 0

        def fsetxattr_cbk(self, frame, cookie, this, op_ret, op_errno, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                print("GLUPY TRACE FSETXATTR CBK- {0:d}: gfid={1:s};  "+
                      "op_ret={2:d}; op_errno={3:d}").format(unique, gfid,
                                                             op_ret, op_errno)
                del self.gfids[key]
                dl.unwind_fsetxattr(frame, cookie, this, op_ret, op_errno,
                                   xdata)
                return 0

        def removexattr_fop(self, frame, this, loc, name, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE REMOVEXATTR FOP- {0:d}:  gfid={1:s}; "+
                      "path={2:s}; name={3:s}").format(unique, gfid,
                                                       loc.contents.path,
                                                       name)
                self.gfids[key] = gfid
                dl.wind_removexattr(frame, POINTER(xlator_t)(), loc, name,
                                    xdata)
                return 0

        def removexattr_cbk(self, frame, cookie, this, op_ret, op_errno,
                            xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                print("GLUPY TRACE REMOVEXATTR CBK- {0:d}: gfid={1:s} "+
                      " op_ret={2:d}; op_errno={3:d}").format(unique, gfid,
                                                              op_ret, op_errno)
                del self.gfids[key]
                dl.unwind_removexattr(frame, cookie, this, op_ret, op_errno,
                                      xdata)
                return 0

        def link_fop(self, frame, this, oldloc, newloc, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                if (newloc.contents.inode):
                        newgfid = uuid2str(newloc.contents.inode.contents.gfid)
                else:
                        newgfid = "0"
                oldgfid = uuid2str(oldloc.contents.inode.contents.gfid)
                print("GLUPY TRACE LINK FOP-{0:d}: oldgfid={1:s}; oldpath={2:s};"+
                      "newgfid={3:s};"+
                      "newpath={4:s}").format(unique, oldgfid,
                                              oldloc.contents.path,
                                              newgfid,
                                              newloc.contents.path)
                self.gfids[key] =  oldgfid
                dl.wind_link(frame, POINTER(xlator_t)(), oldloc, newloc,
                             xdata)
                return 0

        def link_cbk(self, frame, cookie, this, op_ret, op_errno, inode, buf,
                     preparent, postparent, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                if op_ret == 0:
                        statstr = trace_stat2str(buf)
                        preparentstr = trace_stat2str(preparent)
                        postparentstr = trace_stat2str(postparent)
                        print("GLUPY TRACE LINK CBK- {0:d}: op_ret={1:d} "+
                              "*stbuf={2:s}; *prebuf={3:s}; "+
                              "*postbuf={4:s} ").format(unique, op_ret, statstr,
                                                        preparentstr,
                                                        postparentstr)
                else:
                        print("GLUPY TRACE LINK CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d}; "+
                              "op_errno={3:d}").format(unique, gfid,
                                                       op_ret, op_errno)
                del self.gfids[key]
                dl.unwind_link(frame, cookie, this, op_ret, op_errno, inode,
                               buf, preparent, postparent, xdata)
                return 0

        def unlink_fop(self, frame, this, loc, xflag, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE UNLINK FOP- {0:d}; gfid={1:s}; path={2:s}; "+
                      "flag={3:d}").format(unique, gfid, loc.contents.path,
                                           xflag)
                self.gfids[key] = gfid
                dl.wind_unlink(frame, POINTER(xlator_t)(), loc, xflag,
                               xdata)
                return 0

        def unlink_cbk(self, frame, cookie, this, op_ret, op_errno,
                       preparent, postparent, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                if op_ret == 0:
                        preparentstr = trace_stat2str(preparent)
                        postparentstr = trace_stat2str(postparent)
                        print("GLUPY TRACE UNLINK CBK- {0:d}: gfid ={1:s}; "+
                              "op_ret={2:d}; *prebuf={3:s}; "+
                              "*postbuf={4:s} ").format(unique, gfid, op_ret,
                                                        preparentstr,
                                                        postparentstr)
                else:
                        print("GLUPY TRACE UNLINK CBK: {0:d}: gfid ={1:s}; "+
                              "op_ret={2:d}; "+
                              "op_errno={3:d}").format(unique, gfid, op_ret,
                                                       op_errno)
                del self.gfids[key]
                dl.unwind_unlink(frame, cookie, this, op_ret, op_errno,
                                 preparent, postparent, xdata)
                return 0

        def readlink_fop(self, frame, this, loc, size, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE READLINK FOP- {0:d}:  gfid={1:s}; path={2:s};"+
                      " size={3:d}").format(unique, gfid, loc.contents.path,
                                            size)
                self.gfids[key] = gfid
                dl.wind_readlink(frame, POINTER(xlator_t)(), loc, size,
                               xdata)
                return 0

        def readlink_cbk(self, frame, cookie, this, op_ret, op_errno,
                         buf, stbuf, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid  = self.gfids[key]
                if op_ret == 0:
                        statstr = trace_stat2str(stbuf)
                        print("GLUPY TRACE READLINK CBK- {0:d}: gfid={1:s} "+
                              " op_ret={2:d}; op_errno={3:d}; *prebuf={4:s}; "+
                              "*postbuf={5:s} ").format(unique, gfid,
                                                        op_ret, op_errno,
                                                        buf, statstr)
                else:
                        print("GLUPY TRACE READLINK CBK- {0:d}: gfid={1:s} "+
                              " op_ret={2:d}; op_errno={3:d}").format(unique,
                                                                      gfid,
                                                                      op_ret,
                                                                      op_errno)
                del self.gfids[key]
                dl.unwind_readlink(frame, cookie, this, op_ret, op_errno, buf,
                                   stbuf, xdata)
                return 0

        def symlink_fop(self, frame, this, linkpath, loc, umask, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = uuid2str(loc.contents.inode.contents.gfid)
                print("GLUPY TRACE SYMLINK FOP- {0:d}: gfid={1:s}; "+
                      "linkpath={2:s}; path={3:s};"+
                      "umask=0{4:o}").format(unique, gfid, linkpath,
                                             loc.contents.path, umask)
                self.gfids[key] = gfid
                dl.wind_symlink(frame, POINTER(xlator_t)(), linkpath, loc,
                                umask, xdata)
                return 0

        def symlink_cbk(self, frame, cookie, this, op_ret, op_errno,
                        inode, buf, preparent, postparent, xdata):
                unique = dl.get_rootunique(frame)
                key = dl.get_id(frame)
                gfid = self.gfids[key]
                if op_ret == 0:
                        statstr = trace_stat2str(buf)
                        preparentstr = trace_stat2str(preparent)
                        postparentstr = trace_stat2str(postparent)
                        print("GLUPY TRACE SYMLINK CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d}; *stbuf={3:s}; *preparent={4:s}; "+
                              "*postparent={5:s}").format(unique, gfid,
                                                          op_ret, statstr,
                                                          preparentstr,
                                                          postparentstr)
                else:
                        print("GLUPY TRACE SYMLINK CBK- {0:d}: gfid={1:s}; "+
                              "op_ret={2:d}; op_errno={3:d}").format(unique,
                                                                     gfid,
                                                                     op_ret,
                                                                     op_errno)
                del self.gfids[key]
                dl.unwind_symlink(frame, cookie, this, op_ret, op_errno,
                                  inode, buf, preparent, postparent, xdata)
                return 0
